__all__ = ["printh", "write_dset_to_np", "read_dset_from_np"]

import os

import numpy as np
from IPython.display import HTML, display

from riptable import Dataset, Struct


# ------------------------------------------------------------------------------------------
def write_dset_to_np(ds: Dataset, outdir: str, fname: str) -> None:
    """
    Write the columns of a dataset to numpy binary files, one file per column in the specified directory.

    Parameters
    ----------
    ds : Dataset
        A Dataset to write out to disk.
    outdir : str
        The path to the folder where the output will be written.
    fname : str
        The name of the subdirectory to store the columns.

    See Also
    --------
    read_dset_from_np
    """
    os.makedirs(os.path.join(outdir, fname))
    fname = os.path.join(outdir, fname)
    fname = os.path.join(fname, "columns")
    os.makedirs(fname)
    for name, value in ds.items():
        fname_col = os.path.join(fname, str(name))
        np.save(fname_col, value)


def read_dset_from_np(outdir: str, fname: str, mmap: bool = False) -> Dataset:
    """
    Read columns stored as numpy follows to a Dataset.

    Parameters
    ----------
    outdir is the path and fname is the name of the
    subdirectory containing the columns of the dataset
    set mmap = True for memmory mapping. Note this will
    allow quick loading, but has some latency cost elsewhere

    Returns
    -------
    Dataset
        The dataset read in from the specified folder.

    See Also
    --------
    write_dset_to_np
    """
    mmap_mode = None
    if mmap:
        mmap_mode = "r"

    fname = os.path.join(outdir, fname)
    fname = os.path.join(fname, "columns")
    col_dict = dict()
    col_names = os.listdir(fname)
    for i in range(len(col_names)):
        fname_col = os.path.join(fname, col_names[i])
        curr_col_name = col_names[i].replace(".npy", "")
        col_dict[curr_col_name] = np.load(fname_col, mmap_mode=mmap_mode)
    return Dataset(col_dict)


# -----------------------------------------------------------------------------------------
def h5io_to_dataset(io):
    pass


# -----------------------------------------------------------------------------------------
def printh(data):
    """
    Allows jupyter lab/notebook to print multiple HTML renderings in the same output frame.

    Suppose you have three datasets: d1, d2, d3
    In one jupyter cell you could write:
    printh(d1)
    printh(d2)
    printh(d3)
    And all three would be displayed, versus the default, where only the last is shown.
    Will also work for anything else with a _repr_html_ method.

    You can also input a list of elements with _repr_html_ methods so that they display side by side.
    If the jupyter frame isn't wide enough, they'll just display below.

    Parameters
    ----------
    data : object or list of objects
        The object(s) to be rendered for display.
    """
    # multiple items
    if isinstance(data, list):
        html_frames = []
        for i, d in enumerate(data):
            if hasattr(d, "_repr_html_"):
                hfunc = d.__getattr__("_repr_html_")
                if callable(hfunc):
                    html_frames.append(hfunc())
                else:
                    print("_repr_html_ was not callable for item", str(i))
            else:
                print("No _repr_html_ found for item", str(i) + ". Moving to next item.")
        if len(html_frames) > 0:
            master_display = """
                <html>
                <head>
                <style>
                ul.masterlist{
                    list-style-type:none;
                    padding:none;
                }
                ul.masterlist li{
                    margin-left: 30px;
                    display:inline-block;
                    float:left;
                }
                </style>
                </head>
                <body>
                <ul class='masterlist'>
            """
            for d in html_frames:
                master_display += "<li>" + d + "</li>"
            master_display += """
            </ul>
            </body>
            </html>
            """
            display(HTML(master_display))
        else:
            print("No items with _repr_html_ found in list.")

    # single item
    else:
        if hasattr(data, "_repr_html_"):
            hfunc = data.__getattr__("_repr_html_")
            if callable(hfunc):
                display(HTML(hfunc()))
            else:
                print("_repr_html_ was not callable.")
        else:
            print("Input had no _repr_html_ method.")



# __all__ = ['rolling_window', ]

# import numba

## int32 must appear before int64
## float32 signature should appear before float64 signatures
# @numba.guvectorize([
#    'void(int32[:], int32[:], int32[:], int32[:], int32[:], int64, int64, int64)',
#    'void(int32[:], int32[:], int32[:], int64[:], int64[:], int64, int64, int64)',
#    'void(int32[:], int32[:], int32[:], float32[:], float32[:], int64, int64, int64)',
#    'void(int32[:], int32[:], int32[:], float64[:], float64[:], int64, int64, int64)'
#    ], '(m),(n),(n),(o),(o),(),(),()', target='cpu')
# def rolling_window(pGroup, pFirst, pCount, pAccumBin, pSrc, numUnique, totalInputRows, param):

#    # dummy assignment
#    currentSum = pSrc[0]
#    windowSize = param

#    #print('rolling windows', numUnique, totalInputRows, windowSize)

#    for i in range(numUnique):
#        start = pFirst[i]
#        last = start + pCount[i]

#        currentSum = 0

#        # this is the loop for a given group
#        for j in range(start,last):
#            if j >= (start + windowSize):
#                break

#            index = pGroup[j]
#            currentSum  += pSrc[index]
#            pAccumBin[index] = currentSum

#        for j in range(start + windowSize,last):
#            index = pGroup[j]
#            currentSum  += pSrc[index]
#            currentSum  -= pSrc[pGroup[j - windowSize]]
#            pAccumBin[index] = currentSum

# def test_rolling(gb):
#    return gb.calculate_custom_packed(rolling_window, 2)


# @njit(parallel=True)
# def ema_decay(iGroup, nFirstGroup, nCountGroup, time, data, ema, decay):
#    # decay = np.log(2)/(1e3*100)
#    # start on group 1 to skip over 0 bin
#    for i in prange(1, nFirstGroup.shape[0]):
#        start=nFirstGroup[i]
#        last=start + nCountGroup[i]
#        lastEma = 0
#        lastTime = 0
#        for j in range(start,last):
#            index=iGroup[j]
#            lastTime = time[i]-lastTime
#            lastEma = data[j] + lastEma * np.exp(-decay * lastTime)
#        #update output
#        ema[i]=lastEma



# $Id: //Depot/Source/SFW/riptable/Python/core/riptable/tests/runall.py#1 $

import os
import sys
import unittest


def get_all_tests():
    """
    :return:
    """
    tests = (
        fname[:-3]
        for fname in os.listdir(os.path.dirname(__file__) or ".")
        # if (fname.startswith('test_datetime.py') or fname.startswith('test_categorical.py')) and fname.endswith('.py'))
        if fname.startswith("test_") and fname.endswith(".py")
    )
    return sorted(tests)


def run_all(argv, verbosity=3):
    if len(argv) > 1:
        try:
            verbosity = int(argv[1])
        except:
            raise SystemExit("Usage: %s [ verbosity_int (def=%d) ]" % (argv[0], verbosity))
    pkg_name = os.path.basename(os.path.dirname(__file__))
    failures = []
    for test in get_all_tests():
        print("==> Running tests: %s <==\n" % test, file=sys.stderr)
        module_name = "riptable.{pkg_name}.{mod_name}".format(pkg_name=pkg_name, mod_name=test)
        module = __import__(module_name, fromlist=[""])
        utest = unittest.main(module=module, exit=False, argv=[module_name], verbosity=verbosity).result
        if not utest.wasSuccessful():
            failures.append(module_name)
    if failures:
        raise SystemExit("Failure: {}".format(", ".join(failures)))


if __name__ == "__main__":
    run_all(sys.argv)



import unittest

from riptable import *


class PGroupby_Test(unittest.TestCase):
    def test_groupbyhash_cutoffs(self):
        arr = np.random.randint(0, 5, 15)
        arr2 = np.random.randint(0, 5, 15)
        arr3 = hstack([arr, arr2])

        gb_arr = groupbyhash([arr])
        gb_arr2 = groupbyhash([arr2])
        gb_arr3 = groupbyhash([arr3], cutoffs=FA([15, 30], dtype=np.int64))

        ikey = gb_arr3["iKey"]
        ifirst_vals = gb_arr3["iFirstKey"][0]
        ifirst_cutoffs = gb_arr3["iFirstKey"][1]

        self.assertTrue(bool(np.all(gb_arr["iKey"] == ikey[:15])))
        self.assertTrue(bool(np.all(gb_arr["iFirstKey"] == ifirst_vals[: ifirst_cutoffs[0]])))

        self.assertTrue(bool(np.all(gb_arr2["iKey"] == ikey[15:])))
        self.assertTrue(bool(np.all(gb_arr2["iFirstKey"] == ifirst_vals[ifirst_cutoffs[0] :])))
        self.assertEqual(gb_arr3["unique_count"], (gb_arr["unique_count"] + gb_arr2["unique_count"]))

    def test_cutoffs_vs_multikey(self):
        arr = np.random.randint(0, 5, 15)
        arr2 = np.random.randint(0, 5, 15)
        arr3 = hstack([arr, arr2])
        mk_col = hstack([zeros(15), ones(15)])

        gb_mk = groupbyhash([mk_col, arr3])
        gb_arr3 = groupbyhash([arr3], cutoffs=FA([15, 30], dtype=np.int64))
        ikey = gb_arr3["iKey"]
        ifirst_vals = gb_arr3["iFirstKey"][0]
        ifirst_cutoffs = gb_arr3["iFirstKey"][1]

        self.assertTrue(bool(np.all(gb_mk["iFirstKey"][: ifirst_cutoffs[0]] == ifirst_vals[: ifirst_cutoffs[0]])))
        self.assertTrue(bool(np.all(gb_mk["iFirstKey"][ifirst_cutoffs[0] :] == ifirst_vals[ifirst_cutoffs[0] :] + 15)))
        self.assertEqual(gb_mk["unique_count"], gb_arr3["unique_count"])

    def test_single_unique_per_group(self):
        mk_col = hstack([zeros(15), ones(15)])
        pgb = groupbyhash([mk_col], cutoffs=FA([15, 30], dtype=np.int64))
        self.assertTrue(bool(np.all(pgb["iKey"] == 1)))
        self.assertTrue(bool(np.all(pgb["iFirstKey"][0] == 0)))
        self.assertEqual(pgb["unique_count"], 2)

        gb1 = groupbyhash([mk_col[:15]])
        gb2 = groupbyhash([mk_col[15:]])
        self.assertEqual(gb1["unique_count"], 1)
        self.assertEqual(gb2["unique_count"], 1)

    def test_single_cutoff(self):
        arr = np.random.randint(0, 10, 15)
        gb = groupbyhash([arr])
        pgb = groupbyhash([arr], cutoffs=FA([15], dtype=np.int64))

        self.assertTrue(bool(np.all(gb["iKey"] == pgb["iKey"])))
        self.assertTrue(bool(np.all(gb["iFirstKey"] == pgb["iFirstKey"][0])))
        self.assertEqual(len(pgb["iFirstKey"][1]), 1)
        self.assertEqual(pgb["iFirstKey"][1][0], pgb["unique_count"])

    def test_single_item_per_group(self):
        cutoffs = arange(10, dtype=np.int64) + 1
        gb = groupbyhash([arange(10)])
        pgb = groupbyhash([ones(10)], cutoffs=cutoffs)

        self.assertTrue(bool(np.all(gb["iKey"] == pgb["iFirstKey"][1])))
        self.assertEqual(gb["unique_count"], pgb["unique_count"])
        self.assertTrue(bool(np.all(pgb["iKey"] == 1)))
        self.assertTrue(bool(np.all(pgb["iFirstKey"][0] == 0)))

    def test_empty_cutoffs(self):
        a = arange(10)
        cutoffs = a.astype(np.int64) + 1
        f = logical(a % 2)
        no_change = FastArray([0, 1, 1, 2, 2, 3, 3, 4, 4, 5], dtype=np.int64)

        pgb = groupbyhash([ones(10)], filter=f, cutoffs=cutoffs)
        self.assertTrue(bool(np.all(pgb["iKey"] == f)))
        self.assertTrue(bool(np.all(pgb["iFirstKey"][1] == no_change)))


if __name__ == "__main__":
    tester = unittest.main()



import abc
import datetime
from typing import List, Optional

import dateutil.rrule
import pytz
from matplotlib import dates, ticker, units

import riptable as rt

__all__ = []

sizes_ns = {
    "MICROSECONDLY": 1_000,
    "SECONDLY": 1_000_000_000,
    "MINUTELY": 60 * 1_000_000_000,
    "HOURLY": 60 * 60 * 1_000_000_000,
    "DAILY": 24 * 60 * 60 * 1_000_000_000,
    "MONTHLY": 30 * 24 * 60 * 60 * 1_000_000_000,
    "YEARLY": 365 * 24 * 60 * 60 * 1_000_000_000,
}

NANOS_PER_SECOND = 1_000_000_000
MICROS_PER_SECOND = 1_000_000


def dts_to_ns(dts):
    return [dt_to_ns(dt) for dt in dts]


def td_to_ns(td):
    ns = td.days * sizes_ns["DAILY"] + td.seconds * sizes_ns["SECONDLY"] + td.microseconds * 1_000
    return ns


def dt_to_ns(dt):
    return int(dt.timestamp() * NANOS_PER_SECOND)


class DTTicker(abc.ABC):
    tz: pytz.tzfile.DstTzInfo
    interva: int
    start_ns: int
    end_ns: int

    start_utc: datetime.datetime
    start_local: datetime.datetime
    end_utc: datetime.datetime
    end_local: datetime.datetime
    aligned_utc: datetime.datetime
    aligned_local: datetime.datetime

    microsecond_aligned_ticks: List[datetime.datetime]
    millisecond_aligned_ticks: List[datetime.datetime]
    second_aligned_ticks: List[datetime.datetime]
    minute_aligned_ticks: List[datetime.datetime]
    hour_aligned_ticks: List[datetime.datetime]
    day_aligned_ticks: List[datetime.datetime]
    month_aligned_ticks: List[datetime.datetime]
    year_aligned_ticks: List[datetime.datetime]

    ticks_utc: List[datetime.datetime]
    ticks_local: List[datetime.datetime]
    ticks_ns: List[int]

    offset_local: Optional[datetime.datetime]
    tick_format: str
    offset_format: str

    def __init__(self, tz, interval, start_ns, end_ns):
        self.tz = tz
        self.interval = interval
        self.start_ns = start_ns
        self.end_ns = end_ns

        self.make_times()
        self.make_ticks()
        self.find_aligned_ticks()
        self.make_formats()
        return

    def make_times(self):
        self.start_utc = datetime.datetime.fromtimestamp(self.start_ns / NANOS_PER_SECOND, tz=datetime.timezone.utc)
        self.start_local = self.start_utc.astimezone(self.tz)

        self.end_utc = datetime.datetime.fromtimestamp(self.end_ns / NANOS_PER_SECOND, tz=datetime.timezone.utc)
        self.end_local = self.end_utc.astimezone(self.tz)

        self.aligned_local = self.get_aligned()
        self.aligned_utc = self.aligned_local.astimezone(datetime.timezone.utc)
        return

    @abc.abstractmethod
    def get_aligned(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def make_ticks(self):
        raise NotImplementedError()

    def find_aligned_ticks(self):
        self.microsecond_aligned_ticks = [tick for tick in self.ticks_local if tick.microsecond % 1 == 0]
        self.millisecond_aligned_ticks = [
            tick for tick in self.microsecond_aligned_ticks if tick.microsecond % 1_000 == 0
        ]
        self.second_aligned_ticks = [tick for tick in self.millisecond_aligned_ticks if tick.microsecond == 0]
        self.minute_aligned_ticks = [tick for tick in self.second_aligned_ticks if tick.second == 0]
        self.hour_aligned_ticks = [tick for tick in self.minute_aligned_ticks if tick.minute == 0]
        self.day_aligned_ticks = [tick for tick in self.hour_aligned_ticks if tick.hour == 0]
        self.month_aligned_ticks = [tick for tick in self.day_aligned_ticks if tick.day == 0]
        self.year_aligned_ticks = [tick for tick in self.month_aligned_ticks if tick.month == 0]
        return

    @abc.abstractmethod
    def make_formats(self):
        raise NotImplementedError()


class LinTicker(DTTicker):
    step_us = None

    def make_ticks(self):
        pos = self.aligned_utc
        step = datetime.timedelta(microseconds=self.interval * self.step_us)

        ticks = list()
        while pos < self.end_utc:
            if pos >= self.start_utc:
                ticks.append(pos)
            pos += step
        self.ticks_utc = ticks
        self.ticks_local = [tick.astimezone(self.tz) for tick in self.ticks_utc]
        self.ticks_ns = dts_to_ns(self.ticks_utc)
        return


class MicrosecondTicker(LinTicker):
    step_us = 1

    def get_aligned(self):
        aligned = self.tz.localize(self.start_local.replace(microsecond=0, tzinfo=None))
        return aligned

    def make_formats(self):
        if len(self.second_aligned_ticks) == 0:
            first_tick = self.ticks_local[0]
            self.offset_local = first_tick.replace(second=0, microsecond=0)
            self.tick_format = "%fus"
            self.offset_format = "%Y-%m-%d %H:%M:%S"

        elif len(self.second_aligned_ticks) == 1:
            self.offset_local = self.second_aligned_ticks[0]
            self.tick_format = "%fus"
            self.offset_format = "%Y-%m-%d %H:%M:%S"

        else:
            first_tick = self.ticks_local[0]
            self.offset_local = first_tick.replace(hour=0, minute=0, second=0, microsecond=0)
            self.tick_format = "%H:%M:%S.%f"
            self.offset_format = "%Y-%m-%d"
        return


class RRTicker(DTTicker):
    rr: dateutil.rrule.rrule

    freq = None

    def make_ticks(self):
        self.rr = dateutil.rrule.rrule(
            self.freq,
            self.aligned_utc,
            self.interval,
        )
        self.ticks_utc = self.rr.between(self.start_utc, self.end_utc)
        self.ticks_local = [tick.astimezone(self.tz) for tick in self.ticks_utc]
        self.ticks_ns = dts_to_ns(self.ticks_utc)
        return


class SecondTicker(RRTicker):
    freq = dateutil.rrule.SECONDLY

    def get_aligned(self):
        aligned = self.tz.localize(self.start_local.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None))
        return aligned

    def make_formats(self):
        if len(self.day_aligned_ticks) == 0:
            first_tick = self.ticks_local[0]
            self.offset_local = first_tick.replace(hour=0, minute=0, second=0, microsecond=0)
            self.tick_format = "%H:%M:%S"
            self.offset_format = "%Y-%m-%d"
        elif len(self.day_aligned_ticks) == 1:
            self.offset_local = self.day_aligned_ticks[0]
            self.tick_format = "%H:%M:%S"
            self.offset_format = "%Y-%m-%d"
        else:
            self.offset_local = None
            self.tick_format = "%H:%M:%S"
            self.offset_format = "%Y-%m-%d"
        return


class MinuteTicker(RRTicker):
    freq = dateutil.rrule.MINUTELY

    def get_aligned(self):
        aligned = self.tz.localize(self.start_local.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None))
        return aligned

    def make_formats(self):
        if len(self.day_aligned_ticks) == 0:
            first_tick = self.ticks_local[0]
            self.offset_local = first_tick.replace(hour=0, minute=0, second=0, microsecond=0)
            self.tick_format = "%H:%M:%S"
            self.offset_format = "%Y-%m-%d"
        elif len(self.day_aligned_ticks) == 1:
            self.offset_local = self.day_aligned_ticks[0]
            self.tick_format = "%H:%M:%S"
            self.offset_format = "%Y-%m-%d"
        else:
            self.offset_local = None
            self.tick_format = "%H:%M:%S"
            self.offset_format = "%Y-%m-%d"
        return


class HourTicker(RRTicker):
    freq = dateutil.rrule.HOURLY

    def get_aligned(self):
        aligned = self.tz.localize(self.start_local.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None))
        return aligned

    def make_formats(self):
        if len(self.day_aligned_ticks) == 0:
            first_tick = self.ticks_local[0]
            self.offset_local = first_tick.replace(hour=0, minute=0, second=0, microsecond=0)
            self.tick_format = "%H:%M:%S"
            self.offset_format = "%Y-%m-%d"
        elif len(self.day_aligned_ticks) == 1:
            self.offset_local = self.day_aligned_ticks[0]
            self.tick_format = "%H:%M:%S"
            self.offset_format = "%Y-%m-%d"
        else:
            self.offset_local = None
            self.tick_format = "%Y-%m-%d %H:%M:%S"
            self.offset_format = ""
        return


class DayTicker(RRTicker):
    freq = dateutil.rrule.DAILY

    def get_aligned(self):
        aligned = self.tz.localize(
            self.start_local.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0, tzinfo=None)
        )
        return aligned

    def make_formats(self):
        self.offset_local = None
        self.tick_format = "%Y-%m-%d"
        self.offset_format = ""
        return


class MonthTicker(RRTicker):
    freq = dateutil.rrule.MONTHLY

    def get_aligned(self):
        aligned = self.tz.localize(
            self.start_local.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0, tzinfo=None)
        )
        return aligned

    def make_formats(self):
        self.offset_local = None
        self.tick_format = "%Y-%m"
        self.offset_format = ""
        return


class YearTicker(RRTicker):
    freq = dateutil.rrule.YEARLY

    def get_aligned(self):
        aligned = self.tz.localize(
            self.start_local.replace(year=1950, month=1, day=1, hour=0, minute=0, second=0, microsecond=0, tzinfo=None)
        )
        return aligned

    def make_formats(self):
        self.offset_local = None
        self.tick_format = "%Y"
        self.offset_format = ""
        return


class TickFormatter(ticker.Formatter):
    def __init__(self, locator):
        self.locator = locator
        return

    def __call__(self, x, pos=None):
        return self.format_tick(x)

    def format_tick(self, tick_ns):
        tick_utc = datetime.datetime.fromtimestamp(tick_ns / NANOS_PER_SECOND, tz=datetime.timezone.utc)
        tick_local = tick_utc.astimezone(self.locator.ticker.tz)
        name = tick_local.strftime(self.locator.ticker.tick_format)
        if tick_local == self.locator.ticker.offset_local:
            name = "*" + name
        return name

    def format_offset(self):
        if self.locator.ticker.offset_local is None:
            name = ""
        else:
            name = self.locator.ticker.offset_local.strftime(self.locator.ticker.offset_format)
        self.offset_str = name

    def get_offset(self):
        self.format_offset()
        return self.offset_str


class DateTimeNanoLocator(ticker.Locator):
    """
    Locator for axes whose values are given in nanoseconds.
    """

    intervals = {
        "MICROSECONDLY": [1, 2, 5, 10, 25, 50, 100, 250, 1_000, 2_000, 5_000, 10_000, 25_000, 50_000, 100_000, 250_000],
        "SECONDLY": [1, 2, 5, 10, 15, 30],
        "MINUTELY": [1, 2, 5, 10, 15, 30],
        "HOURLY": [1, 3, 4, 6, 12],
        "DAILY": [1, 2, 3, 7, 14, 21],
        "MONTHLY": [1, 2, 3, 4, 6],
        "YEARLY": [1, 2, 4, 5, 10, 20, 40, 50],
    }

    max_ticks = {
        "MICROSECONDLY": 10,
        "SECONDLY": 10,
        "MINUTELY": 10,
        "HOURLY": 10,
        "DAILY": 10,
        "MONTHLY": 10,
        "YEARLY": 10,
    }

    def __init__(self, tz):
        self.scale = None
        self.interval = None
        self.tz = tz
        return

    def __call__(self):
        (vmin, vmax) = self.axis.get_view_interval()
        self.ticks = self.tick_values(vmin, vmax)
        return self.ticks

    def tick_values(self, vmin, vmax):
        # Compute how wide the view is.
        width_nanos = abs(vmax - vmin)
        (self.scale, self.interval) = self.get_tick_interval(width_nanos)
        ticks = self.get_ticks(vmin, vmax)
        return ticks

    def get_tick_interval(self, width):
        for (scale, sizes) in self.intervals.items():
            nanos = sizes_ns[scale]
            max_tick = self.max_ticks[scale]
            for size in sizes:
                if width / (nanos * size) < max_tick:
                    return (scale, size)
        else:
            raise ValueError("Could not determine tick interval")

    def get_ticks(self, start, end):
        start = int(start)
        end = int(end)
        if self.scale == "MICROSECONDLY":
            self.ticker = MicrosecondTicker(self.tz, self.interval, start, end)
        elif self.scale == "SECONDLY":
            self.ticker = SecondTicker(self.tz, self.interval, start, end)
        elif self.scale == "MINUTELY":
            self.ticker = MinuteTicker(self.tz, self.interval, start, end)
        elif self.scale == "HOURLY":
            self.ticker = HourTicker(self.tz, self.interval, start, end)
        elif self.scale == "DAILY":
            self.ticker = DayTicker(self.tz, self.interval, start, end)
        elif self.scale == "MONTHLY":
            self.ticker = MonthTicker(self.tz, self.interval, start, end)
        elif self.scale == "YEARLY":
            self.ticker = YearTicker(self.tz, self.interval, start, end)
        else:
            raise ValueError("Scale not recognized.")

        return self.ticker.ticks_ns


class DateTimeNanoScalarConverter(units.ConversionInterface):
    @staticmethod
    def default_units(x, axis):
        return ("ns", x._timezone._to_tz)

    @staticmethod
    def axisinfo(unit, axis):
        (u, rt_tz) = unit
        if rt_tz == "NYC":
            tz = pytz.timezone("America/New_York")
        else:
            tz = pytz.utc

        majloc = DateTimeNanoLocator(tz)
        majfmt = TickFormatter(majloc)
        info = units.AxisInfo(
            majloc=majloc,
            majfmt=majfmt,
            label="DateTime",
        )
        return info


class DateTimeScalarConverter(units.ConversionInterface):
    @staticmethod
    def default_units(x, axis):
        return "date"

    @staticmethod
    def axisinfo(unit, axis):
        majloc = dates.AutoDateLocator()
        majfmt = dates.AutoDateFormatter(majloc)
        info = units.AxisInfo(
            majloc=majloc,
            majfmt=majfmt,
            label="Date",
        )
        return info


class TSTicker(abc.ABC):
    step_us = None

    def __init__(
        self,
        interval,
        start_ns,
        end_ns,
    ):
        self.interval = interval
        self.start_ns = start_ns
        self.end_ns = end_ns

        self.make_time_deltas()
        self.make_ticks()
        self.find_aligned_ticks()
        self.make_formats()
        return

    def make_time_deltas(self):
        self.start_td = datetime.timedelta(microseconds=self.start_ns / 1_000)
        self.end_td = datetime.timedelta(microseconds=self.end_ns / 1_000)
        self.aligned_td = self.get_aligned()
        return

    @abc.abstractmethod
    def get_aligned(self):
        raise NotImplementedError()

    def make_ticks(self):
        pos = self.aligned_td
        step = datetime.timedelta(microseconds=self.interval * self.step_us)
        ticks = list()
        while pos < self.end_td:
            if pos > self.start_td:
                ticks.append(pos)
            pos += step
        self.ticks_td = ticks
        self.ticks_ns = [td_to_ns(td) for td in self.ticks_td]

    def find_aligned_ticks(self):
        self.microsecond_aligned_ticks = [tick for tick in self.ticks_td if tick.microseconds % 1 == 0]
        self.millisecond_aligned_ticks = [
            tick for tick in self.microsecond_aligned_ticks if tick.microseconds % 1_000 == 0
        ]
        self.second_aligned_ticks = [tick for tick in self.millisecond_aligned_ticks if tick.microseconds == 0]
        self.minute_aligned_ticks = [tick for tick in self.second_aligned_ticks if tick.seconds % 60 == 0]
        self.hour_aligned_ticks = [tick for tick in self.minute_aligned_ticks if tick.seconds % 3600 == 0]
        self.day_aligned_ticks = [tick for tick in self.hour_aligned_ticks if tick.seconds == 0]

    @abc.abstractmethod
    def make_formats(self):
        raise NotImplementedError()


class MicrosecondTSTicker(TSTicker):
    step_us = 1

    def get_aligned(self):
        aligned = self.start_td - datetime.timedelta(microseconds=self.start_td.microseconds)
        return aligned

    def make_formats(self):
        if len(self.day_aligned_ticks) == 0:
            first_tick = self.ticks_td[0]
            self.offset = first_tick - datetime.timedelta(microseconds=first_tick.microseconds)
            self.tick_format = "{F:06}us"
            self.offset_format = "{D} days, {H:02}:{M:02}:{S:02}"

        elif len(self.day_aligned_ticks) == 1:
            self.offset = self.day_aligned_ticks[0]
            self.tick_format = "{F:06}us"
            self.offset_format = "{D} days, {H:02}:{M:02}:{S:02}"

        else:
            self.offset = None
            self.tick_format = "{D} days, {H:02}:{M:02}:{S:02}"
            self.offset_format = ""
        return


class SecondTSTicker(TSTicker):
    step_us = 1_000_000

    def get_aligned(self):
        aligned = self.start_td - datetime.timedelta(
            seconds=self.start_td.seconds, microseconds=self.start_td.microseconds
        )
        return aligned

    def make_formats(self):
        if len(self.day_aligned_ticks) == 0:
            first_tick = self.ticks_td[0]
            self.offset = first_tick - datetime.timedelta(first_tick.seconds)
            self.tick_format = "{H:02}:{M:02}:{S:02}"
            self.offset_format = "{D} days"

        elif len(self.day_aligned_ticks) == 1:
            self.offset = self.day_aligned_ticks[0]
            self.tick_format = "{H:02}:{M:02}:{S:02}"
            self.offset_format = "{D} days"

        else:
            self.offset = None
            self.tick_format = "{D} days, {H:02}:{M:02}:{S:02}"
            self.offset_format = ""
        return


class MinuteTSTicker(TSTicker):
    step_us = 60 * 1_000_000

    def get_aligned(self):
        aligned = self.start_td - datetime.timedelta(
            seconds=self.start_td.seconds, microseconds=self.start_td.microseconds
        )
        return aligned

    def make_formats(self):
        if len(self.day_aligned_ticks) == 0:
            first_tick = self.ticks_td[0]
            self.offset = first_tick - datetime.timedelta(
                seconds=first_tick.seconds, microseconds=first_tick.microseconds
            )
            self.tick_format = "{H:02}:{M:02}:{S:02}"
            self.offset_format = "{D} days"

        elif len(self.day_aligned_ticks) == 1:
            self.offset = self.day_aligned_ticks[0]
            self.tick_format = "{H:02}:{M:02}:{S:02}"
            self.offset_format = "{D} days"

        else:
            self.offset = None
            self.tick_format = "{D} days, {H:02}:{M:02}:{S:02}"
            self.offset_format = ""
        return


class HourTSTicker(TSTicker):
    step_us = 60 * 60 * 1_000_000

    def get_aligned(self):
        aligned = self.start_td - datetime.timedelta(
            seconds=self.start_td.seconds, microseconds=self.start_td.microseconds
        )
        return aligned

    def make_formats(self):
        if len(self.day_aligned_ticks) == 0:
            first_tick = self.ticks_td[0]
            self.offset = first_tick - datetime.timedelta(
                seconds=first_tick.seconds, microseconds=first_tick.microseconds
            )
            self.tick_format = "{H:02}:{M:02}:{S:02}"
            self.offset_format = "{D} days"

        elif len(self.day_aligned_ticks) == 1:
            self.offset = self.day_aligned_ticks[0]
            self.tick_format = "{H:02}:{M:02}:{S:02}"
            self.offset_format = "{D} days"

        else:
            self.offset = None
            self.tick_format = "{D} days, {H:02}:{M:02}:{S:02}"
            self.offset_format = ""
        return


class DayTSTicker(TSTicker):
    step_us = 24 * 60 * 60 * 1_000_000

    def get_aligned(self):
        aligned = self.start_td - datetime.timedelta(
            seconds=self.start_td.seconds, microseconds=self.start_td.microseconds
        )
        return aligned

    def make_formats(self):
        self.offset = None
        self.tick_format = "{D} days"
        self.offset_format = ""
        return


class TimeSpanLocator(ticker.Locator):
    intervals = {
        "MICROSECONDLY": [1, 2, 5, 10, 25, 50, 100, 250, 1_000, 2_000, 5_000, 10_000, 25_000, 50_000, 100_000, 250_000],
        "SECONDLY": [1, 2, 5, 10, 15, 30],
        "MINUTELY": [1, 2, 5, 10, 15, 30],
        "HOURLY": [1, 3, 4, 6, 12],
        "DAILY": [1, 2, 3, 7, 14, 21, 30, 60, 90, 180, 365, 730, 3650],
    }

    max_ticks = {
        "MICROSECONDLY": 10,
        "SECONDLY": 10,
        "MINUTELY": 10,
        "HOURLY": 10,
        "DAILY": 10,
    }

    def __init__(self):
        self.scale = None
        self.interval = None
        return

    def __call__(self):
        (vmin, vmax) = self.axis.get_view_interval()
        self.ticks = self.tick_values(vmin, vmax)
        return self.ticks

    def tick_values(self, vmin, vmax):
        # Compute how wide the view is.
        width_nanos = abs(vmax - vmin)
        (self.scale, self.interval) = self.get_tick_interval(width_nanos)
        ticks = self.get_ticks(vmin, vmax)
        return ticks

    def get_tick_interval(self, width):
        for (scale, sizes) in self.intervals.items():
            nanos = sizes_ns[scale]
            max_tick = self.max_ticks[scale]
            for size in sizes:
                if width / (nanos * size) < max_tick:
                    return (scale, size)
        else:
            raise ValueError("Could not determine tick interval")

    def get_ticks(self, start, end):
        start = int(start)
        end = int(end)
        if self.scale == "MICROSECONDLY":
            self.ticker = MicrosecondTSTicker(self.interval, start, end)
        elif self.scale == "SECONDLY":
            self.ticker = SecondTSTicker(self.interval, start, end)
        elif self.scale == "MINUTELY":
            self.ticker = MinuteTSTicker(self.interval, start, end)
        elif self.scale == "HOURLY":
            self.ticker = HourTSTicker(self.interval, start, end)
        elif self.scale == "DAILY":
            self.ticker = DayTSTicker(self.interval, start, end)
        else:
            raise ValueError("Scale not recognized.")

        return self.ticker.ticks_ns


def td_to_dhmsf(td):
    d = td.days
    sec_rem = td.seconds
    (h, sec_rem) = divmod(sec_rem, 3600)
    (m, s) = divmod(sec_rem, 60)
    f = td.microseconds
    return (d, h, m, s, f)


class TimeSpanFormatter(ticker.Formatter):
    def __init__(self, locator):
        self.locator = locator
        return

    def __call__(self, x, pos=None):
        return self.format_tick(x)

    def format_tick(self, tick_ns):
        tick_td = datetime.timedelta(microseconds=tick_ns / 1_000)
        (days, hours, minutes, seconds, microseconds) = td_to_dhmsf(tick_td)
        name = self.locator.ticker.tick_format.format(
            D=days,
            H=hours,
            M=minutes,
            S=seconds,
            F=microseconds,
        )
        if tick_td == self.locator.ticker.offset:
            name = "*" + name
        return name

    def format_offset(self):
        if self.locator.ticker.offset is None:
            name = ""
        else:
            offset = self.locator.ticker.offset
            (days, hours, minutes, seconds, microseconds) = td_to_dhmsf(offset)
            name = self.locator.ticker.offset_format.format(
                D=days,
                H=hours,
                M=minutes,
                S=seconds,
                F=microseconds,
            )
        self.offset_str = name

    def get_offset(self):
        self.format_offset()
        return self.offset_str


class TimeSpanConverter(units.ConversionInterface):
    @staticmethod
    def default_units(x, axis):
        return "ns"

    @staticmethod
    def axisinfo(unit, axis):
        majloc = TimeSpanLocator()
        majfmt = TimeSpanFormatter(majloc)
        info = units.AxisInfo(
            majloc=majloc,
            majfmt=majfmt,
            label="TimeSpan",
        )
        return info


units.registry[rt.DateTimeNanoScalar] = DateTimeNanoScalarConverter
units.registry[rt.DateScalar] = DateTimeScalarConverter
units.registry[rt.TimeSpanScalar] = TimeSpanConverter



"""
General-purpose settings for configuring riptable behavior.

This module provides a class encapsulating riptable settings and
feature flags, along with functions for retrieving a top-level,
process-wide instance of the class.
"""
__all__ = ["Settings", "get_global_settings"]

from typing import NamedTuple


# TODO: Make this a frozen @dataclass once riptable only supports Python 3.7+.
#       Once the change is made, we can also implement a function which looks for environment
#       variables (e.g. "RIPTABLE_ENABLE_NUMBA_CACHE", "RIPTABLE_MAX_NUM_THREADS") specifying
#       overrides for default values, then creates the global Settings instance accordingly.
class Settings(NamedTuple):
    """
    Encapsulates process-wide settings and feature-flags for riptable.
    """

    enable_numba_cache: bool = False
    """
    Controls whether the numba JIT cache is enabled for functions within riptable.
    This is disabled (False) by default because the caching can lead to occasional
    segfaults in numba-compiled code for some users, possibly caused by a race condition
    or filesystem non-atomicity.
    """


__global_settings = Settings()
"""Global (process-wide) `Settings` instance."""


def get_global_settings() -> Settings:
    """
    Get the global (process-wide) `Settings` instance.

    Returns
    -------
    Settings
        Global (process-wide) settings.
    """
    return __global_settings



import unittest
import pytest
import itertools
from numpy.random import default_rng
from riptable import *
import riptable as rt
from riptable.rt_numpy import QUANTILE_METHOD_NP_KW, gb_np_quantile, np_rolling_nanquantile

from numpy.testing import (
    assert_array_equal,
    assert_almost_equal,
    assert_array_almost_equal,
)


class GroupByOps_Test(unittest.TestCase):
    def test_categorical_values(self):
        vals = np.random.choice(["a", "b", "c", "d", "e"], 50)
        ds = Dataset({"vals": vals})
        # TODO what assertions need to be done on 'c' and 'gb'
        c = Categorical(vals, ordered=False)
        gb = ds.gb("vals")

        vals = np.random.choice(["a", "b", "c", "d", "e"], 50)
        ds = Dataset({"vals": vals})
        c = Categorical(vals)
        gb = ds.gb("vals", lex=True)
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = FA(vals, unicode=True)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=False, unicode=True)
        gb = ds.gb("vals")
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = np.random.randint(0, 1000, 50)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=False)
        gb = ds.gb("vals")
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = np.random.choice([1.0, 2.0, 3.0, 4.0, 5.0, 6.0], 50)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=False)
        gb = ds.gb("vals")
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = np.random.choice(["a", "b", "c", "d", "e"], 50)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=False, sort_gb=False)
        gb = ds.gb("vals")
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = FA(vals, unicode=True)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=False, unicode=True, sort_gb=False)
        gb = ds.gb("vals")
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = np.random.randint(0, 1000, 50)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=False, sort_gb=False)
        gb = ds.gb("vals")
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = np.random.choice([1.0, 2.0, 3.0, 4.0, 5.0, 6.0], 50)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=False, sort_gb=False)
        gb = ds.gb("vals")
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        # 4/29/2019 - iFirstKey gets generated later, uses a new routine
        # need to set ordered to False to correctly match groupby results
        # groupby always uses first-occurence
        # or ordered=True groupby(lex=True)

        vals = np.random.choice(["a", "b", "c", "d", "e"], 50)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=True)
        gb = ds.gb("vals", lex=True)
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = FA(vals, unicode=True)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, unicode=True, ordered=True)
        gb = ds.gb("vals", lex=True)
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = np.random.randint(0, 1000, 50)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=True)
        gb = ds.gb("vals", lex=True)
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

        vals = np.random.choice([1.0, 2.0, 3.0, 4.0, 5.0, 6.0], 50)
        ds = Dataset({"vals": vals})
        c = Categorical(vals, ordered=True)
        gb = ds.gb("vals", lex=True)
        self.assertTrue(bool(np.all(c.ifirstkey == gb.ifirstkey)))

    def test_categorical_values_cats(self):
        pass

    def test_categorical_codes_cats(self):
        pass

    def test_categorical_mapping(self):
        pass

    def test_categorical_multikey(self):
        pass

    def test_gb_userfunc(self):
        ds = Dataset({"strs": np.random.choice(["a", "b", "c"], 20)})
        ds.data = arange(20)
        # make sure np.quantile stlye works where second argument is passed
        ds.gb("strs").apply_nonreduce(np.quantile, 0.5)

    def test_gb_extra_data(self):
        ds = Dataset({"strs": np.random.choice(["a", "b", "c"], 20)})
        ds.col_AAA = ones(20)
        ds.col_BBB = np.random.rand(20)
        extra = arange(20)
        extra.set_name("extra")
        ds2 = ds.gb("strs").sum(extra)

        final_cols = ["strs", "col_AAA", "col_BBB", "extra"]
        self.assertTrue(bool(np.all(list(ds2) == final_cols)))

    def test_cat_add_dataset(self):
        ds = Dataset({"strs": np.random.choice(["a", "b", "c"], 20)})
        ds.col_AAA = ones(20)
        ds.col_BBB = np.random.rand(20)
        c = Categorical(ds.strs)
        ds2 = c.sum(dataset=ds)
        final_cols = ["strs", "col_AAA", "col_BBB"]
        self.assertTrue(bool(np.all(list(ds2) == final_cols)))

        extra = arange(20)
        extra.set_name("extra")
        ds3 = c.sum(extra, dataset=ds)
        final_cols = ["strs", "col_AAA", "col_BBB", "extra"]
        self.assertTrue(bool(np.all(list(ds3) == final_cols)))

    def test_warn_conflicting_data(self):
        ds = Dataset({"strs": np.random.choice(["a", "b", "c"], 20)})
        ds.col_AAA = ones(20)
        ds.col_BBB = np.random.rand(20)
        c = Categorical(np.random.choice(["a", "b", "c"], 20))
        with self.assertWarns(UserWarning):
            ds2 = c.sum(ds, dataset=ds)

    def test_error_no_data(self):
        c = Categorical(np.random.choice(["a", "b", "c"], 20))
        with self.assertRaises(ValueError):
            _ = c.sum()

        ds = Dataset({"strs": np.random.choice(["a", "b", "c"], 20)})
        ds.col_AAA = ones(20)
        ds.col_BBB = np.random.rand(20)
        gb = ds.gb("strs")
        gb._dataset = None
        with self.assertRaises(ValueError):
            _ = gb.sum()

    def test_ema_funcs(self):
        ds = Dataset({"test": arange(10), "group2": arange(10) % 3})

        gb = ds.gb("group2")
        c = Categorical(ds.group2)

        # decay_rate now required to be filled in
        gb_result = gb.ema_normal(time=arange(10), decay_rate=1.0)
        c_result = c.ema_normal(ds.test, time=arange(10), decay_rate=1.0)

        for k, col in gb_result.items():
            self.assertTrue(bool(np.all(col == c_result[k])))

    def test_agg_funcs(self):
        ds = Dataset({"col_" + str(i): np.random.rand(20) for i in range(10)})
        ds.keycol = np.random.choice(["a", "b", "c"], 20)
        c = Categorical(ds.keycol)
        flt = np.random.rand(20).view(FA)
        flt.set_name("flt")

        gb_result = ds.gb("keycol").agg({"col_0": ["sum", np.mean], "col_5": ["min", np.var], "flt": "max"}, flt)
        c_result = c.agg(
            {"col_0": ["sum", np.mean], "col_5": ["min", np.var], "flt": "max"},
            flt,
            dataset=ds[:-1],
        )

        for dsname, d_total in gb_result.items():
            c_total = c_result[dsname]
            for k, v in d_total.items():
                self.assertTrue(bool(np.all(c_total[k] == v)))

    def test_agg_funcs_np(self):
        c = Categorical(np.random.choice(["a", "b", "c"], 20))

        rt_result = c.agg(["min", "max"], arange(20))
        np_result = c.agg([np.amin, np.amax], arange(20))

        for dsname, d_total in rt_result.items():
            c_total = np_result[dsname]
            for k, v in d_total.items():
                self.assertTrue(bool(np.all(c_total[k] == v)))

    def test_agg_funcs_error(self):
        c = Categorical(np.random.choice(["a", "b", "c"], 20))
        with self.assertRaises(ValueError):
            _ = c.agg({"badname": "min"}, arange(20))

        with self.assertRaises(ValueError):
            _ = c.agg(["min", "max", "badfunc"], arange(20))

    def test_groups_order(self):
        strings = np.array(["b", "a", "a", "z", "y", "b", "c"])
        ds = Dataset(
            {
                "col1": Categorical(strings, ordered=True),
                "col2": Categorical(strings, ordered=False),
                "col3": strings,
            }
        )
        d1 = ds.col1.groups
        d2 = ds.col2.groups
        d3 = ds.gb("col3").groups
        d4 = ds.gbu("col3").groups

        for colname, idx in d1.items():
            for d in (d2, d3, d4):
                self.assertTrue(bool(np.all(d[colname] == idx)))


def quantile_params_generator(dtypes, max_N, windowed, with_cats, with_nans, with_infs, seed, scrunity_level=1):

    rng = default_rng(seed)

    for dtype in dtypes:

        # N = 1 case
        N = 1
        N_cat = 1
        window = 1
        for quantile in [round(x, 4) for x in rng.random(3 * scrunity_level)]:
            curr_frac_options = [(0, 0, 0)]
            if with_nans:
                curr_frac_options += [(1, 0, 0)]
            if with_infs:
                curr_frac_options += [(0, 1, 0), (0, 0, 1)]
            for nan_fraction, inf_fraction, minus_inf_fraction in curr_frac_options:
                yield N, N_cat, window, quantile, nan_fraction, inf_fraction, minus_inf_fraction, dtype

        frac_lists = [
            [0.0, 1] + [round(x, 4) for x in rng.random(1 * scrunity_level)],
            [0.0, 1] + [round(x, 4) for x in rng.random(1 * scrunity_level)],
            [0.0, 1] + [round(x, 4) for x in rng.random(1 * scrunity_level)],
        ]

        frac_options = [element for element in itertools.product(*frac_lists) if sum(element) <= 1]
        frac_options_only_nans = [(x, 0, 0) for x in frac_lists[0]]

        curr_frac_options = [(0, 0, 0)]
        if with_nans:
            curr_frac_options = frac_options_only_nans
        if with_infs:
            curr_frac_options = frac_options

        if with_cats:
            # N_cat = 1 case
            N = rng.integers(400, 600)
            N_cat = 1
            window = rng.integers(100, 200)
            for quantile in [round(x, 3) for x in rng.random(2 * scrunity_level)]:
                for nan_fraction, inf_fraction, minus_inf_fraction in curr_frac_options:
                    yield N, N_cat, window, quantile, nan_fraction, inf_fraction, minus_inf_fraction, dtype

        # window = 1 case
        N = rng.integers(400, 600)
        N_cat = rng.integers(30, 50)
        window = 1
        for quantile in [round(x, 3) for x in rng.random(2 * scrunity_level)]:
            for nan_fraction, inf_fraction, minus_inf_fraction in curr_frac_options:
                yield N, N_cat, window, quantile, nan_fraction, inf_fraction, minus_inf_fraction, dtype

        # window > N case
        N = rng.integers(400, 600)
        N_cat = rng.integers(30, 50)
        window = rng.integers(700, 900)
        for quantile in [round(x, 3) for x in rng.random(1 * scrunity_level)]:
            for nan_fraction, inf_fraction, minus_inf_fraction in curr_frac_options:
                yield N, N_cat, window, quantile, nan_fraction, inf_fraction, minus_inf_fraction, dtype

        # quantile = 0., 1. cases
        N = rng.integers(400, 600)
        N_cat = rng.integers(10, 40)
        window_list = [1, 3, rng.integers(10, 40)]
        if not windowed:
            window_list = [1]
        for window in window_list:
            for quantile in [0.0, 1.0]:
                for nan_fraction, inf_fraction, minus_inf_fraction in curr_frac_options:
                    yield N, N_cat, window, quantile, nan_fraction, inf_fraction, minus_inf_fraction, dtype

        for N in (
            rng.integers(20, 1000, size=3 * scrunity_level).tolist()
            + rng.integers(10_000, 50_000, size=2 * scrunity_level).tolist()
        ):
            if N > max_N:
                continue
            window_list = rng.integers(15, N, size=1 * scrunity_level)
            if not windowed:
                window_list = [1]
            for window in window_list:
                cat_list = rng.integers(2, N, size=1 * scrunity_level)
                if not with_cats:
                    cat_list = [1]
                for N_cat in cat_list:
                    for quantile in [round(x, 3) for x in rng.random(2 * scrunity_level)]:
                        for nan_fraction, inf_fraction, minus_inf_fraction in curr_frac_options:
                            yield N, N_cat, window, quantile, nan_fraction, inf_fraction, minus_inf_fraction, dtype


def fills_params_generator():
    float_types = [np.float32, np.float64]
    integral_types = [
        np.int8,
        np.uint8,
        np.int16,
        np.uint16,
        np.int32,
        np.uint32,
        np.int64,
        np.uint64,
    ]

    all_types = float_types + integral_types

    common_types = [np.int64, np.float64]

    # N = 1 case
    N = 1
    N_cat = 1
    limit = 0
    fill_values = [None, 0, 15]
    nan_fractions = [0.5, 0.95]
    types = common_types

    for fill_val in fill_values:
        for nan_fraction in nan_fractions:
            yield N, N_cat, nan_fraction, types, limit, fill_val

    # types
    N = 1_000
    N_cat = 200
    limits = [0, 2]
    nan_fraction = 0.5
    types = all_types
    fill_val = None
    for limit in limits:
        yield N, N_cat, nan_fraction, types, limit, fill_val

    # limits
    N = 1_020
    N_cat = 210
    limits = [0, 1, 2, 4, 6, 10]
    nan_fractions = [0.5, 0.95]
    types = common_types
    fill_val = None
    for limit in limits:
        for nan_fraction in nan_fractions:
            yield N, N_cat, nan_fraction, types, limit, fill_val

    # fill_vals
    N = 965
    N_cat = 300
    limits = [0, 2]
    nan_fraction = 0.5
    types = common_types
    fill_values = [None, 0, -15, 100]
    for limit in limits:
        for fill_val in fill_values:
            yield N, N_cat, nan_fraction, types, limit, fill_val


class TestGoupByOps_functions:
    @pytest.mark.parametrize("N", [1, 2, 5, 100, 3467, 60_000, 120_000])
    @pytest.mark.parametrize("N_cat", [1, 2, 3, 10, 1_000, 20_000])
    @pytest.mark.parametrize("nan_fraction", [0.0, 0.1, 0.3, 0.7, 0.9, 0.9999999, 1.0])
    @pytest.mark.parametrize("rep", [1])
    def test_groupbyops_all_min_max(self, N, N_cat, nan_fraction, rep):
        this_seed = (N * N_cat * rep * int(nan_fraction * 1000)) % 10_000
        rng = default_rng(this_seed)

        ds = Dataset({"categ": rng.integers(0, N_cat, size=N)})
        ds["categ"] = Cat(ds["categ"])
        categories = ds.categ._grouping.uniquelist[0]
        ds.data = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.data[rng.random(size=N) < nan_fraction] = np.nan

        ds.int_data = rng.integers(-10_000, 10_000, size=N)

        int_nan_idxs = rng.random(size=N) < nan_fraction
        ds.int_data[int_nan_idxs] = np.nan

        # transform into float and then into np.array to have proper np.nans there
        np_int_data_copy = ds.int_data.astype(np.float64)._np
        np_int_data_copy[int_nan_idxs] = np.nan

        cols = ["data", "int_data"]

        dsmin = ds.categ.min(ds)
        dsmax = ds.categ.max(ds)

        dsnanmin = ds.categ.nanmin(ds)
        dsnanmax = ds.categ.nanmax(ds)

        for column in cols:
            # check that the result on the whole dataset equals to the results on
            # individual columns
            assert_array_equal(dsmin[column], ds.categ.min(ds[column])[column])
            assert_array_equal(dsmax[column], ds.categ.max(ds[column])[column])

            assert_array_equal(dsnanmin[column], ds.categ.nanmin(ds[column])[column])
            assert_array_equal(dsnanmax[column], ds.categ.nanmax(ds[column])[column])

        def err_msg():
            return f"""
N = {N}
N_cat = {N_cat}
nan_fraction = {nan_fraction}
column = {column}
category = {category}"""

        for column in cols:
            dtype = ds[column].dtype
            for category in rng.choice(categories, min(len(categories), 60)):
                data_res_min = dsmin[column][dsmin.categ == category]
                data_res_max = dsmax[column][dsmax.categ == category]
                data_res_nanmin = dsnanmin[column][dsnanmin.categ == category]
                data_res_nanmax = dsnanmax[column][dsnanmax.categ == category]
                if dtype in [np.int32, np.int64]:
                    curr_slice = np_int_data_copy[ds.categ == category]
                else:
                    curr_slice = ds.filter(ds.categ == category)[column]._np
                data_res_min_np = np.min(curr_slice)
                data_res_max_np = np.max(curr_slice)
                data_res_nanmin_np = np.nanmin(curr_slice)
                data_res_nanmax_np = np.nanmax(curr_slice)
                # make answer float to transform integer invalids into np.nans for comparison
                assert_almost_equal(data_res_min.astype(float)[0], data_res_min_np, err_msg=err_msg())
                assert_almost_equal(data_res_max.astype(float)[0], data_res_max_np, err_msg=err_msg())
                assert_almost_equal(data_res_nanmin.astype(float)[0], data_res_nanmin_np, err_msg=err_msg())
                assert_almost_equal(data_res_nanmax.astype(float)[0], data_res_nanmax_np, err_msg=err_msg())

    @pytest.mark.parametrize(
        "N, N_cat, window, q, nan_fraction, inf_fraction, minus_inf_fraction, dtype",
        quantile_params_generator(
            [np.int32, np.float64],
            max_N=100_000,
            windowed=False,
            with_cats=True,
            with_nans=True,
            with_infs=False,
            seed=100,
        ),
    )
    def test_groupbyops_quantile(self, N, N_cat, window, q, nan_fraction, inf_fraction, minus_inf_fraction, dtype):
        this_seed = (N * N_cat * int(q * 1000) * int(nan_fraction * 1000)) % 10_000
        rng = default_rng(this_seed)
        kwargs = {QUANTILE_METHOD_NP_KW: "midpoint"}

        def err_msg():
            return f"""
q = {q}
N = {N}
N_cat = {N_cat}
nan_fraction = {nan_fraction}
dtype = {dtype}"""

        ds = Dataset({"categ": rng.integers(0, N_cat, size=N)})
        categories = np.unique(ds.categ)
        ds["categ"] = Cat(ds["categ"])

        ds.data = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N).astype(dtype)
        ds.data2 = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N).astype(dtype)

        nan_idxs = rng.random(size=N) < nan_fraction
        # need to convert to float for numpy to understand nans.
        # np_data_copy will only be used for numpy operations for checks.
        np_data_copy = ds.data.copy().astype(np.float64)
        np_data_copy[nan_idxs] = np.nan
        ds.data[nan_idxs] = np.nan

        data_quant = ds.categ.quantile(ds.data, q=q)
        data2_quant = ds.categ.quantile(ds.data2, q=q)
        both_quant = ds.categ.quantile((ds.data, ds.data2), q=q)

        assert_array_almost_equal(data_quant.data, both_quant.data, err_msg=err_msg())
        assert_array_almost_equal(data2_quant.data2, both_quant.data2, err_msg=err_msg())

        for category in rng.choice(categories, min(len(categories), 15)):
            curr_filter = ds.categ == category
            curr_slice = ds.filter(curr_filter)
            data_res = data_quant.filter(data_quant.categ == category).data
            # converting this to float for nanmax and nanmean.
            if q * (1 - q) == 0:
                data_res = data_res.astype(np.float64)
            data_res_np = np.quantile(np_data_copy[curr_filter], q=q, **kwargs)
            assert_almost_equal(data_res, data_res_np, err_msg=err_msg())
            data2_res = data2_quant.filter(data2_quant.categ == category).data2
            data2_res_np = np.quantile(curr_slice.data2, q=q, **kwargs)
            assert_almost_equal(data2_res, data2_res_np, err_msg=err_msg())

        both_percentiles = ds.categ.percentile((ds.data, ds.data2), q=q * 100)
        assert_array_almost_equal(both_quant.data, both_percentiles.data, err_msg=err_msg())
        assert_array_almost_equal(both_quant.data2, both_percentiles.data2, err_msg=err_msg())

        # nanquantile
        data_nanquant = ds.categ.nanquantile(ds.data, q=q)
        data2_nanquant = ds.categ.nanquantile(ds.data2, q=q)
        both_nanquant = ds.categ.nanquantile((ds.data, ds.data2), q=q)

        assert_array_almost_equal(data2_nanquant.data2, data2_quant.data2, err_msg=err_msg())
        assert_array_almost_equal(data_nanquant.data, both_nanquant.data, err_msg=err_msg())
        assert_array_almost_equal(data2_nanquant.data2, both_nanquant.data2, err_msg=err_msg())

        for category in rng.choice(categories, min(len(categories), 15)):
            curr_filter = ds.categ == category
            curr_slice = ds.filter(curr_filter)
            data_res = data_nanquant.filter(data_nanquant.categ == category).data
            # converting this to float for nanmax and nanmean.
            if q * (1 - q) == 0:
                data_res = data_res.astype(np.float64)
            data_res_np = np.nanquantile(np_data_copy[curr_filter], q=q, **kwargs)
            assert_almost_equal(data_res, data_res_np, err_msg=err_msg())
            data2_res = data2_nanquant.filter(data2_nanquant.categ == category).data2
            data2_res_np = np.nanquantile(curr_slice.data2, q=q, **kwargs)
            assert_almost_equal(data2_res, data2_res_np, err_msg=err_msg())

        both_nanpercentiles = ds.categ.nanpercentile((ds.data, ds.data2), q=q * 100)
        assert_array_almost_equal(both_nanquant.data, both_nanpercentiles.data, err_msg=err_msg())
        assert_array_almost_equal(both_nanquant.data2, both_nanpercentiles.data2, err_msg=err_msg())

    @pytest.mark.parametrize(
        "N, N_cat, window, q, nan_fraction, inf_fraction, minus_inf_fraction, dtype",
        quantile_params_generator(
            [np.float64], max_N=100_000, windowed=False, with_cats=True, with_nans=True, with_infs=True, seed=200
        ),
    )
    def test_groupbyops_quantile_infs(self, N, N_cat, window, q, nan_fraction, inf_fraction, minus_inf_fraction, dtype):
        def err_msg():
            return f"""
q = {q}
N = {N}
N_cat = {N_cat}
nan_fraction = {nan_fraction}
inf_fraction = {inf_fraction}
minus_inf_fraction = {minus_inf_fraction}"""

        this_seed = (N * N_cat * int(q * 1000) * int(nan_fraction * 1000)) % 10_000
        rng = default_rng(this_seed)
        ds = Dataset({"categ": rng.integers(0, N_cat, size=N)})
        categories = np.unique(ds.categ)
        ds["categ"] = Cat(ds["categ"])

        # infs and integers don't work normally, only check floats
        ds.data = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.data2 = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)

        # notice that np.nanquantile doesn't handle infs properly
        # use gb_np_quantile for comparison

        rand_idxs = rng.random(size=N)
        ds.data[rand_idxs < (nan_fraction + inf_fraction + minus_inf_fraction)] = -np.inf
        ds.data[rand_idxs < (nan_fraction + inf_fraction)] = np.inf
        ds.data[rand_idxs < (nan_fraction)] = np.nan

        data_quant = ds.categ.quantile(ds.data, q=q)
        data2_quant = ds.categ.quantile(ds.data2, q=q)
        both_quant = ds.categ.quantile((ds.data, ds.data2), q=q)

        assert_array_almost_equal(data_quant.data, both_quant.data, err_msg=err_msg())
        assert_array_almost_equal(data2_quant.data2, both_quant.data2, err_msg=err_msg())

        for category in rng.choice(categories, min(len(categories), 15)):
            curr_filter = ds.categ == category
            curr_slice = ds.filter(curr_filter)
            data_res = data_quant.filter(data_quant.categ == category).data
            data_res_np = gb_np_quantile(ds.data[curr_filter], q=q, is_nan_function=False)
            assert_almost_equal(data_res, data_res_np, err_msg=err_msg())
            data2_res = data2_quant.filter(data2_quant.categ == category).data2
            data2_res_np = gb_np_quantile(curr_slice.data2, q=q, is_nan_function=False)
            assert_almost_equal(data2_res, data2_res_np, err_msg=err_msg())

        both_percentiles = ds.categ.percentile((ds.data, ds.data2), q=q * 100)
        assert_array_almost_equal(both_quant.data, both_percentiles.data, err_msg=err_msg())
        assert_array_almost_equal(both_quant.data2, both_percentiles.data2, err_msg=err_msg())

        # nanquantile
        data_nanquant = ds.categ.nanquantile(ds.data, q=q)
        data2_nanquant = ds.categ.nanquantile(ds.data2, q=q)
        both_nanquant = ds.categ.nanquantile((ds.data, ds.data2), q=q)

        assert_array_almost_equal(data2_nanquant.data2, data2_quant.data2, err_msg=err_msg())
        assert_array_almost_equal(data_nanquant.data, both_nanquant.data, err_msg=err_msg())
        assert_array_almost_equal(data2_nanquant.data2, both_nanquant.data2, err_msg=err_msg())

        for category in rng.choice(categories, min(len(categories), 15)):
            curr_filter = ds.categ == category
            curr_slice = ds.filter(curr_filter)
            data_res = data_nanquant.filter(data_nanquant.categ == category).data
            data_res_np = gb_np_quantile(ds.data[curr_filter], q=q, is_nan_function=True)
            assert_almost_equal(data_res, data_res_np, err_msg=err_msg())
            data2_res = data2_nanquant.filter(data2_nanquant.categ == category).data2
            data2_res_np = gb_np_quantile(curr_slice.data2, q=q, is_nan_function=True)
            assert_almost_equal(data2_res, data2_res_np, err_msg=err_msg())

        both_nanpercentiles = ds.categ.nanpercentile((ds.data, ds.data2), q=q * 100)
        assert_array_almost_equal(both_nanquant.data, both_nanpercentiles.data, err_msg=err_msg())
        assert_array_almost_equal(both_nanquant.data2, both_nanpercentiles.data2, err_msg=err_msg())

    @pytest.mark.parametrize("N", [1, 300])
    @pytest.mark.parametrize("N_cat", [20])
    @pytest.mark.parametrize("q", [[0.0, 0.34, 0.81], [0.0, 0.5, 1.0, 0.2], [0.123, 0.12345, 0.1], 0.734])
    def test_groupbyops_quantile_multi_params(self, N, N_cat, q):
        def err_msg():
            return f"""
q = {q}
N = {N}
N_cat = {N_cat}"""

        this_seed = N * N_cat * int(np.sum(q) * 1000) % 10_000
        rng = default_rng(this_seed)

        ds = Dataset({"categ": rng.integers(0, N_cat, size=N)})
        categories = np.unique(ds.categ)
        ds["categ"] = Cat(ds["categ"])

        ds.data = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.data2 = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.col_0 = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.col_1 = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.int_data = rng.integers(-10_000, 10_000, size=N)
        ds.int_data2 = rng.integers(0, 100, size=N)

        cols = ["data", "data2", "col_0", "col_1", "int_data", "int_data2"]

        quant_ds = ds.categ.quantile(ds, q=q)
        quant_sep = ds.categ.quantile((ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=q)
        percentile_sep = ds.categ.percentile(
            (ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=np.array(q) * 100
        )
        nanpercentile_sep = ds.categ.nanpercentile(
            (ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=np.array(q) * 100
        )
        nanquantile_sep = ds.categ.nanquantile((ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=q)

        assert_array_equal(quant_ds.imatrix_make(), quant_sep.imatrix_make())
        assert_array_equal(quant_ds.imatrix_make(), percentile_sep.imatrix_make())
        assert_array_equal(quant_ds.imatrix_make(), nanpercentile_sep.imatrix_make())
        assert_array_equal(quant_ds.imatrix_make(), nanquantile_sep.imatrix_make())

        q_cols = quant_ds.keys()

        if np.isscalar(q):
            q = [q]
            q_cols = cols

        # columns are ordered like [col + quantile for quantile in q for col in cols] if len(q) > 1

        for i_col, column in enumerate(cols):
            for i_q, quantile in enumerate(q):
                quant_col = q_cols[i_col * len(q) + i_q]
                for category in rng.choice(categories, min(len(categories), 15)):
                    curr_filter = ds.categ == category
                    curr_slice = ds.filter(curr_filter)
                    data_res = quant_ds[quant_col][quant_ds.categ == category]
                    data_res_np = gb_np_quantile(curr_slice[column], q=quantile, is_nan_function=False)
                    assert_almost_equal(data_res, data_res_np, err_msg=err_msg() + f"\n{column} {quant_col}\n{q_cols}")

        # multi-key

        ds.cat2 = rng.integers(0, N_cat, size=N)
        del ds.categ
        ds.categ = rng.integers(0, N_cat, size=N)
        mk_cat = Cat([ds.categ, ds.cat2])
        categories = np.unique(mk_cat._fa)
        fa_cat_aligned = mk_cat.first(mk_cat._fa)["col_0"]

        quant_ds = mk_cat.quantile(ds, q=q)
        quant_sep = mk_cat.quantile((ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=q)
        percentile_sep = mk_cat.percentile(
            (ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=np.array(q) * 100
        )
        nanpercentile_sep = mk_cat.nanpercentile(
            (ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=np.array(q) * 100
        )
        nanquantile_sep = mk_cat.nanquantile((ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=q)

        assert_array_almost_equal(quant_ds.imatrix_make(), quant_sep.imatrix_make())
        assert_array_almost_equal(quant_ds.imatrix_make(), percentile_sep.imatrix_make())
        assert_array_almost_equal(quant_ds.imatrix_make(), nanpercentile_sep.imatrix_make())
        assert_array_almost_equal(quant_ds.imatrix_make(), nanquantile_sep.imatrix_make())

        q_cols = quant_ds.keys()

        if np.isscalar(q):
            q = [q]
            q_cols = cols

        # columns are ordered like [col + quantile for quantile in q for col in cols] if len(q) > 1

        for i_col, column in enumerate(cols):
            for i_q, quantile in enumerate(q):
                quant_col = q_cols[i_col * len(q) + i_q]
                for category in rng.choice(categories, min(len(categories), 15)):
                    curr_filter = mk_cat._fa == category
                    curr_slice = ds.filter(curr_filter)
                    data_res = quant_ds[quant_col][fa_cat_aligned == category]
                    data_res_np = gb_np_quantile(curr_slice[column], q=quantile, is_nan_function=False)
                    assert_array_almost_equal(
                        data_res, data_res_np, err_msg=err_msg() + f"\n{column} {quant_col}\n{q_cols}"
                    )

    @pytest.mark.parametrize("N", [1, 10_000, 100_000])
    @pytest.mark.parametrize("N_cat", [1, 100])
    @pytest.mark.parametrize("dtype", [np.int32, np.int64, np.float32, np.float64])
    def test_groupbyops_basic_functions(self, N, N_cat, dtype):
        this_seed = N * N_cat
        rng = default_rng(this_seed)

        def err_msg():
            return f"""
N = {N}
N_cat = {N_cat}
this_seed = {this_seed}
dtype = {dtype}
gb_function = {gb_function}
np_function = {np_function}"""

        ds = Dataset({"categ": rng.integers(0, N_cat, size=N)})
        categories = np.unique(ds.categ)
        ds["categ"] = Cat(ds["categ"])

        ds.data = rng.uniform(-100, 100, size=N).astype(dtype)

        if dtype in [np.int32, np.int64]:
            np_data_copy = np.array(ds.data.copy().astype(np.float64))
        else:
            np_data_copy = np.array(ds.data.copy())

        # np_data_copy[nan_idxs] = np.nan
        # ds.data[nan_idxs] = np.nan

        gb_functions = [
            ds.categ.sum,
            ds.categ.mean,
            ds.categ.var,
            ds.categ.std,
            ds.categ.nansum,
            ds.categ.nanmean,
            ds.categ.nanvar,
            ds.categ.nanstd,
        ]
        np_functions = [
            np.sum,
            np.mean,
            lambda x: np.var(x, ddof=1),
            lambda x: np.std(x, ddof=1),
            np.nansum,
            np.nanmean,
            lambda x: np.nanvar(x, ddof=1),
            lambda x: np.nanstd(x, ddof=1),
        ]

        def test_categories(ds, gb_result):
            for category in rng.choice(categories, min(len(categories), 7)):
                curr_filter = ds.categ == category
                if not curr_filter.sum():
                    return
                np_result = np_function(np_data_copy[curr_filter])

                if dtype in [np.int32, np.int64]:
                    categ_result = gb_result.data.astype(np.float64)[gb_result.categ == category][0]
                else:
                    categ_result = gb_result.data[gb_result.categ == category][0]

                if dtype not in [np.float32]:
                    assert_almost_equal(categ_result, np_result, err_msg=err_msg())
                else:
                    decimal = 2 if N <= 10_000 else 1
                    assert_almost_equal(categ_result, np_result, err_msg=err_msg(), decimal=decimal)

        for phase in range(2):
            if phase == 1:
                nan_idxs = rng.random(N) < 0.5
                ds.data[nan_idxs] = np.nan
                np_data_copy[nan_idxs] = np.nan

            for gb_function, np_function in zip(gb_functions, np_functions):
                gb_result = gb_function(ds.data)
                test_categories(ds, gb_result)

    @pytest.mark.parametrize(
        "N, N_cat, window, q, nan_fraction, inf_fraction, minus_inf_fraction, dtype",
        quantile_params_generator(
            [np.int32, np.float64],
            max_N=1_000,
            windowed=True,
            with_cats=True,
            with_nans=True,
            with_infs=False,
            seed=500,
        ),
    )
    def test_groupbyops_rolling_quantile(
        self, N, N_cat, window, q, nan_fraction, inf_fraction, minus_inf_fraction, dtype
    ):
        this_seed = (N * N_cat * int(q * 1000) * int(nan_fraction * 1000)) % 10_000
        rng = default_rng(this_seed)

        def err_msg():
            return f"""
N = {N}
N_cat = {N_cat}
window = {window}
q = {q}
nan_fraction = {nan_fraction}
dtype = {dtype}"""

        ds = Dataset({"categ": rng.integers(0, N_cat, size=N)})
        categories = np.unique(ds.categ)
        ds["categ"] = Cat(ds["categ"])

        ds.data = rng.normal(rng.uniform(-100, 100), rng.uniform(0.00, 100), size=N).astype(dtype)
        ds.data2 = rng.normal(rng.uniform(-100, 100), rng.uniform(0.00, 100), size=N).astype(dtype)

        nan_idxs = rng.random(size=N) < nan_fraction
        # need to convert to float for numpy to understand nans.
        # np_data_copy will only be used for numpy operations for checks.
        np_data_copy = ds.data.copy().astype(np.float64)
        np_data_copy[nan_idxs] = np.nan
        ds.data[nan_idxs] = np.nan

        np_data2_copy = ds.data2.copy().astype(np.float64)
        np_data2_copy[nan_idxs] = np.nan
        ds.data2[nan_idxs] = np.nan

        data_quant = ds.categ.rolling_quantile(ds.data, q=q, window=window)
        data2_quant = ds.categ.rolling_quantile(ds.data2, q=q, window=window)
        both_quant = ds.categ.rolling_quantile((ds.data, ds.data2), q=q, window=window)
        gb_quant = ds.groupby("categ").rolling_quantile(q=q, window=window)

        assert_array_almost_equal(data_quant.data, both_quant.data, err_msg=err_msg())
        assert_array_almost_equal(data2_quant.data2, both_quant.data2, err_msg=err_msg())
        assert_array_almost_equal(data_quant.data, gb_quant.data, err_msg=err_msg())

        for category in rng.choice(categories, min(len(categories), 15)):
            curr_filter = ds.categ == category
            curr_window = min(window, curr_filter.sum())
            if curr_window == 0:
                continue

            data_res = data_quant.filter(curr_filter).data
            curr_np_data = np.array(np_data_copy[curr_filter])
            data_res_np = np_rolling_nanquantile(a=curr_np_data, q=q, window=curr_window)
            assert_almost_equal(data_res[curr_window - 1 :], data_res_np, err_msg=err_msg())

            data2_res = data2_quant.filter(curr_filter).data2
            curr_np_data2 = np.array(np_data2_copy[curr_filter])
            data2_res_np = np_rolling_nanquantile(a=curr_np_data2, q=q, window=curr_window)
            assert_almost_equal(data2_res[curr_window - 1 :], data2_res_np, err_msg=err_msg())

    @pytest.mark.parametrize(
        "N, N_cat, window, q, nan_fraction, inf_fraction, minus_inf_fraction, dtype",
        quantile_params_generator(
            [np.float64], max_N=1_000, windowed=True, with_cats=True, with_nans=True, with_infs=True, seed=600
        ),
    )
    def test_groupbyops_rolling_quantile_infs(
        self, N, N_cat, window, q, nan_fraction, inf_fraction, minus_inf_fraction, dtype
    ):
        def err_msg():
            return f"""
N = {N}
N_cat = {N_cat}
q = {q}
window = {window}
nan_fraction = {nan_fraction}
inf_fraction = {inf_fraction}
minus_inf_fraction = {minus_inf_fraction}"""

        this_seed = (N * N_cat * int(q * 1000) * int(nan_fraction * 1000)) % 10_000
        rng = default_rng(this_seed)

        ds = Dataset({"categ": rng.integers(0, N_cat, size=N)})
        categories = np.unique(ds.categ)
        ds["categ"] = Cat(ds["categ"])

        data_quant = Dataset()
        # check that works fine with multiple data columns
        for i in range(6):
            ds[f"data{i}"] = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)

            rand_idxs = rng.random(size=N)
            ds[f"data{i}"][rand_idxs < (nan_fraction + inf_fraction + minus_inf_fraction)] = -np.inf
            ds[f"data{i}"][rand_idxs < (nan_fraction + inf_fraction)] = np.inf
            ds[f"data{i}"][rand_idxs < (nan_fraction)] = np.nan

            data_quant[f"data{i}"] = ds.categ.rolling_quantile(ds[f"data{i}"], q=q, window=window)[f"data{i}"]

        all_quant = ds.categ.rolling_quantile([ds[f"data{i}"] for i in range(6)], q=q, window=window)
        gb_quant = ds.groupby("categ").rolling_quantile(q=q, window=window)

        for i in range(6):
            assert_array_almost_equal(data_quant[f"data{i}"], all_quant[f"data{i}"], err_msg=err_msg())
            assert_array_almost_equal(data_quant[f"data{i}"], gb_quant[f"data{i}"], err_msg=err_msg())

        for i in range(6):
            for category in rng.choice(categories, min(len(categories), 1)):
                curr_filter = ds.categ == category
                curr_window = min(window, curr_filter.sum())
                if curr_window == 0:
                    continue

                data_res = data_quant.filter(curr_filter)[f"data{i}"]
                curr_np_data = np.array(ds[f"data{i}"][curr_filter])
                data_res_np = np_rolling_nanquantile(a=curr_np_data, q=q, window=curr_window)
                assert_almost_equal(data_res[curr_window - 1 :], data_res_np, err_msg=err_msg())

    @pytest.mark.parametrize("N", [300])
    @pytest.mark.parametrize("N_cat", [10])
    @pytest.mark.parametrize("q", [[0.0, 0.34, 0.81], [0.0, 0.5, 1.0, 0.24], [0.123, 0.314, 0.1], 0.734])
    @pytest.mark.parametrize("window", [20])
    def test_groupbyops_rolling_quantile_multi_params(self, N, N_cat, q, window):
        def err_msg():
            return f"""
N = {N}
N_cat = {N_cat}
q = {q}
window = {window}"""

        this_seed = N * N_cat * int(np.sum(q) * 1000) % 10_000
        rng = default_rng(this_seed)

        ds = Dataset({"categ": rng.integers(0, N_cat, size=N)})
        categories = np.unique(ds.categ)
        ds["categ"] = Cat(ds["categ"])

        ds.data = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.data2 = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.col_0 = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.col_1 = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
        ds.int_data = rng.integers(-10_000, 10_000, size=N)
        ds.int_data2 = rng.integers(0, 100, size=N)

        cols = ["data", "data2", "col_0", "col_1", "int_data", "int_data2"]

        quant_ds = ds.categ.rolling_quantile(ds, q=q, window=window)
        quant_sep = ds.categ.rolling_quantile(
            (ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=q, window=window
        )

        assert_array_equal(quant_ds.imatrix_make(), quant_sep.imatrix_make())

        q_cols = quant_ds.keys()

        if np.isscalar(q):
            q = [q]
            q_cols = cols

        # columns are ordered like [col + quantile for quantile in q for col in cols] if len(q) > 1

        for i_col, column in enumerate(cols):
            for i_q, quantile in enumerate(q):
                quant_col = q_cols[i_col * len(q) + i_q]
                for category in rng.choice(categories, min(len(categories), 15)):
                    curr_filter = ds.categ == category
                    curr_window = min(window, curr_filter.sum())
                    if curr_window == 0:
                        continue
                    data_res = quant_ds.filter(curr_filter)[quant_col]
                    curr_np_data = np.array(ds[column][curr_filter])
                    data_res_np = np_rolling_nanquantile(a=curr_np_data, q=quantile, window=curr_window)
                    assert_almost_equal(data_res[curr_window - 1 :], data_res_np, err_msg=err_msg())

        # multi-key

        ds.cat2 = rng.integers(0, N_cat, size=N)
        del ds.categ
        ds.categ = rng.integers(0, N_cat, size=N)
        mk_cat = Cat([ds.categ, ds.cat2])
        categories = np.unique(mk_cat._fa)
        fa_cat_aligned = mk_cat.first(mk_cat._fa)["col_0"]

        quant_ds = mk_cat.rolling_quantile(ds, q=q, window=window)
        quant_sep = mk_cat.rolling_quantile(
            (ds.data, ds.data2, ds.col_0, ds.col_1, ds.int_data, ds.int_data2), q=q, window=window
        )

        assert_array_almost_equal(quant_ds.imatrix_make(), quant_sep.imatrix_make())

        q_cols = quant_ds.keys()

        if np.isscalar(q):
            q = [q]
            q_cols = cols

        for i_col, column in enumerate(cols):
            for i_q, quantile in enumerate(q):
                quant_col = q_cols[i_col * len(q) + i_q]
                for category in rng.choice(categories, min(len(categories), 15)):
                    curr_filter = mk_cat._fa == category
                    curr_window = min(window, curr_filter.sum())
                    if curr_window == 0:
                        continue
                    data_res = quant_ds[quant_col][curr_filter]
                    curr_np_data = np.array(ds[column][curr_filter])
                    data_res_np = np_rolling_nanquantile(a=curr_np_data, q=quantile, window=curr_window)
                    assert_almost_equal(data_res[curr_window - 1 :], data_res_np, err_msg=err_msg())

    @pytest.mark.skip("Old fill functions were replaced by numba implementations")
    @pytest.mark.parametrize("N", [1, 100, 1_000])
    @pytest.mark.parametrize("N_cat", [1, 10, 300])
    @pytest.mark.parametrize("limit", [0, 3, 10])
    @pytest.mark.parametrize("nan_fraction", [0.5, 0.95])
    @pytest.mark.parametrize("fill_value", [0, -15, 100, None])
    def test_groupbyops_compare_fill_nb_fill(self, N, N_cat, limit, nan_fraction, fill_value):
        def err_msg():
            return f"""
N = {N}
N_cat = {N_cat}
limit = {limit}"""

        this_seed = N * N_cat * np.sum(limit) % 10_000
        rng = default_rng(this_seed)

        ds = rt.Dataset()
        ds.categ = rt.Cat(rng.integers(0, N_cat, size=N))
        ds.categ2 = rng.integers(0, N_cat, size=N)
        multi_cat = rt.Cat([ds["categ"], ds["categ2"]])

        ds.gb_key1 = rng.integers(0, N_cat, size=N)
        ds.gb_key2 = rng.integers(0, N_cat, size=N)
        single_gb = ds.groupby("gb_key1")
        multi_gb = ds.groupby(["gb_key1", "gb_key2"])

        for data_type in [
            np.int8,
            np.uint8,
            np.int16,
            np.uint16,
            np.int32,
            np.uint32,
            np.int64,
            np.uint64,
            np.float32,
            np.float64,
        ]:
            temp_data = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
            if type in [np.uint8, np.uint16, np.uint32, np.uint64]:
                temp_data = np.abs(temp_data)

            ds[f"data_{data_type.__name__}"] = rt.FA(temp_data).astype(data_type)
            ds[f"data_{data_type.__name__}"][rng.random(N) < nan_fraction] = np.nan

        # Cat

        old_cat_fill_f = ds.categ.fill_forward(ds, limit=limit, fill_val=fill_value)
        new_cat_fill_f = ds.categ.nb_fill_forward(ds, limit=limit, fill_val=fill_value)

        old_multi_cat_fill_f = multi_cat.fill_forward(ds, limit=limit, fill_val=fill_value)
        new_multi_cat_fill_f = multi_cat.nb_fill_forward(ds, limit=limit, fill_val=fill_value)

        assert old_cat_fill_f.equals(new_cat_fill_f, exact=True)
        assert old_multi_cat_fill_f.equals(new_multi_cat_fill_f)

        old_cat_fill_b = ds.categ.fill_backward(ds, limit=limit, fill_val=fill_value)
        new_cat_fill_b = ds.categ.nb_fill_backward(ds, limit=limit, fill_val=fill_value)

        old_multi_cat_fill_b = multi_cat.fill_backward(ds, limit=limit, fill_val=fill_value)
        # because of weirdness of names when doing operations on multi-key categoricals
        # exact=True always fails for multi-key categoricals if the keys have the same name
        old_multi_cat_fill_b.col_rename("gb_key_1", "gb_key_0")
        new_multi_cat_fill_b = multi_cat.nb_fill_backward(ds, limit=limit, fill_val=fill_value)

        assert old_cat_fill_b.equals(new_cat_fill_b, exact=True)
        assert old_multi_cat_fill_b.equals(new_multi_cat_fill_b)

        # GB

        old_gb_fill_f = single_gb.fill_forward(limit=limit, fill_val=fill_value)
        new_gb_fill_f = single_gb.nb_fill_forward(limit=limit, fill_val=fill_value)

        old_multi_gb_fill_f = multi_gb.fill_forward(limit=limit, fill_val=fill_value)
        new_multi_gb_fill_f = multi_gb.nb_fill_forward(limit=limit, fill_val=fill_value)

        assert old_gb_fill_f.equals(new_gb_fill_f, exact=True)
        assert old_multi_gb_fill_f.equals(new_multi_gb_fill_f, exact=True)

        old_gb_fill_b = single_gb.fill_backward(limit=limit, fill_val=fill_value)
        new_gb_fill_b = single_gb.nb_fill_backward(limit=limit, fill_val=fill_value)

        old_multi_gb_fill_b = multi_gb.fill_backward(limit=limit, fill_val=fill_value)
        new_multi_gb_fill_b = multi_gb.nb_fill_backward(limit=limit, fill_val=fill_value)

        assert old_gb_fill_b.equals(new_gb_fill_b, exact=True)
        assert old_multi_gb_fill_b.equals(new_multi_gb_fill_b, exact=True)

        # Make sure names are the same for multi-key categoricals
        ds = rt.Dataset()
        ds.categ = rt.Cat(rng.integers(0, N_cat, size=N))
        ds.categ2 = rng.integers(0, N_cat, size=N)
        multi_cat = rt.Cat([ds["categ"], ds["categ2"]])

        for data_type in [
            np.int64,
            np.float64,
        ]:
            temp_data = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
            if type in [np.uint8, np.uint16, np.uint32, np.uint64]:
                temp_data = np.abs(temp_data)

            ds[f"data_{data_type.__name__}"] = rt.FA(temp_data).astype(data_type)
            ds[f"data_{data_type.__name__}"][rng.random(N) < nan_fraction] = np.nan

        # make copy of a categorical, as multiple applications of method to a categorical changes names of columns
        multi_cat2 = rt.Cat([ds["categ"], ds["categ2"]])

        # old
        old_multi_cat_fill_f = multi_cat.fill_forward(ds, limit=limit, fill_val=fill_value)
        old_multi_cat_fill_f_second_call = multi_cat.fill_forward(ds, limit=limit, fill_val=fill_value)

        # new
        new_multi_cat_fill_f = multi_cat2.fill_forward(ds, limit=limit, fill_val=fill_value)
        new_multi_cat_fill_f_second_call = multi_cat2.fill_forward(ds, limit=limit, fill_val=fill_value)

        # just check that column names are exactly the same, we checked that the values are equal in tests above
        assert old_multi_cat_fill_f.keys() == new_multi_cat_fill_f.keys()
        assert old_multi_cat_fill_f_second_call.keys() == new_multi_cat_fill_f_second_call.keys()

    @pytest.mark.parametrize(
        "N, N_cat, nan_fraction, types, limit, fill_val",
        fills_params_generator(),
    )
    def test_groupbyops_fills(self, N, N_cat, nan_fraction, types, limit, fill_val):

        this_seed = N * N_cat * int(nan_fraction * 1000) % 10_000
        rng = default_rng(this_seed)

        def err_msg(**kwargs):
            base = f"""
seed = {this_seed}
N = {N}
N_cat = {N_cat}
nan_fraction = {nan_fraction}
limit = {limit}
fill_val = {fill_val}"""
            for k, v in kwargs.items():
                base += f"""
{k} = {v}
"""
            return base

        ds = rt.Dataset()
        ds.categ = rt.Cat(rng.integers(0, N_cat, size=N))
        ds.categ2 = rng.integers(0, N_cat, size=N)
        multi_cat = rt.Cat([ds["categ"], ds["categ2"]])

        ds.gb_key1 = rng.integers(0, N_cat, size=N)
        ds.gb_key2 = rng.integers(0, N_cat, size=N)
        single_gb = ds.groupby("gb_key1")
        multi_gb = ds.groupby(["gb_key1", "gb_key2"])

        data_column_names = []

        for data_type in types:
            temp_data = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
            if type in [np.uint8, np.uint16, np.uint32, np.uint64]:
                temp_data = np.abs(temp_data)

            col_name = f"data_{data_type.__name__}"
            ds[col_name] = rt.FA(temp_data).astype(data_type)
            ds[col_name][rng.random(N) < nan_fraction] = np.nan
            data_column_names.append(col_name)

        # Cat

        cat_ff = ds.categ.fill_forward(ds, limit=limit, fill_val=fill_val)
        cat_fb = ds.categ.fill_backward(ds, limit=limit, fill_val=fill_val)

        single_cat_result = {"ff": cat_ff, "fb": cat_fb}

        multi_cat_ff = multi_cat.fill_forward(ds, limit=limit, fill_val=fill_val)
        multi_cat_fb = multi_cat.fill_backward(ds, limit=limit, fill_val=fill_val)

        multi_cat_result = {"ff": multi_cat_ff, "fb": multi_cat_fb}

        for categorical, ds_result in zip([ds.categ, multi_cat], [single_cat_result, multi_cat_result]):
            unique_categories = np.unique(categorical._fa)
            for category in rng.choice(unique_categories, min(len(unique_categories), 8)):
                for col in data_column_names:
                    curr_filter = categorical._fa == category
                    curr_data = ds[col][curr_filter]
                    fa_ff = curr_data.fill_forward(limit=limit, fill_val=fill_val)
                    fa_fb = curr_data.fill_backward(limit=limit, fill_val=fill_val)

                    ds_ff = ds_result["ff"][col][curr_filter]
                    ds_fb = ds_result["fb"][col][curr_filter]

                    assert_array_equal(ds_ff, fa_ff, err_msg=err_msg(curr_data=curr_data))
                    assert_array_equal(ds_fb, fa_fb, err_msg=err_msg(curr_data=curr_data))

        # GB

        gb_ff = single_gb.fill_forward(limit=limit, fill_val=fill_val)
        gb_fb = single_gb.fill_backward(limit=limit, fill_val=fill_val)

        single_gb_result = {"ff": gb_ff, "fb": gb_fb}

        multi_cat_ff = multi_gb.fill_forward(limit=limit, fill_val=fill_val)
        multi_cat_fb = multi_gb.fill_backward(limit=limit, fill_val=fill_val)

        multi_gb_result = {"ff": multi_cat_ff, "fb": multi_cat_fb}

        for gb, ds_result in zip([single_gb, multi_gb], [single_gb_result, multi_gb_result]):
            unique_categories = np.unique(gb.grouping._iKey.unique())
            for category in rng.choice(unique_categories, min(len(unique_categories), 8)):
                for col in data_column_names:
                    curr_filter = gb.grouping._iKey == category
                    curr_data = ds[col][curr_filter]
                    fa_ff = curr_data.fill_forward(limit=limit, fill_val=fill_val)
                    fa_fb = curr_data.fill_backward(limit=limit, fill_val=fill_val)

                    ds_ff = ds_result["ff"][col][curr_filter]
                    ds_fb = ds_result["fb"][col][curr_filter]

                    assert_array_equal(ds_ff, fa_ff)
                    assert_array_equal(ds_fb, fa_fb)

    def test_groupbyops_fills_inplace(self):
        N = 1234
        N_cat = 56
        types = [
            np.float32,
            np.float64,
            np.int8,
            np.uint8,
            np.int16,
            np.uint16,
            np.int32,
            np.uint32,
            np.int64,
            np.uint64,
        ]
        nan_fraction = 0.5

        this_seed = N * N_cat * int(nan_fraction * 1000) % 10_000
        rng = default_rng(this_seed)

        ds = rt.Dataset()
        ds.categ = rt.Cat(rng.integers(0, N_cat, size=N))
        ds.categ2 = rng.integers(0, N_cat, size=N)
        multi_cat = rt.Cat([ds["categ"], ds["categ2"]])

        ds.gb_key1 = rng.integers(0, N_cat, size=N)
        ds.gb_key2 = rng.integers(0, N_cat, size=N)
        single_gb = ds.groupby("gb_key1")
        multi_gb = ds.groupby(["gb_key1", "gb_key2"])

        data_column_names = []

        for data_type in types:
            temp_data = rng.normal(rng.uniform(-10_000, 10_000), rng.uniform(0.00, 10_000), size=N)
            if type in [np.uint8, np.uint16, np.uint32, np.uint64]:
                temp_data = np.abs(temp_data)

            col_name = f"data_{data_type.__name__}"
            ds[col_name] = rt.FA(temp_data).astype(data_type)
            ds[col_name][rng.random(N) < nan_fraction] = np.nan
            data_column_names.append(col_name)

        ds_copy = ds.copy(deep=True)

        for cat_object in [ds.categ, multi_cat]:
            # not inplace
            ds = ds_copy.copy(deep=True)
            ds_ff = cat_object.fill_forward(ds[data_column_names])
            assert ds.equals(ds_copy)

            # inplace
            cat_object.fill_forward(ds[data_column_names], inplace=True)
            assert not ds.equals(ds_copy)
            assert (ds[data_column_names]).equals(ds_ff[data_column_names])

        # not inplace
        ds = ds_copy.copy(deep=True)
        single_gb = ds.groupby("gb_key1")
        ds_ff = single_gb.fill_forward()
        assert ds.equals(ds_copy)

        # inplace
        single_gb.fill_forward(inplace=True)
        assert not ds.equals(ds_copy)
        assert (ds[data_column_names]).equals(ds_ff[data_column_names])

        # not inplace
        ds = ds_copy.copy(deep=True)
        multi_gb = ds.groupby(["gb_key1", "gb_key2"])

        ds_ff = multi_gb.fill_forward()
        assert ds.equals(ds_copy)

        # inplace
        multi_gb.fill_forward(inplace=True)
        assert not ds.equals(ds_copy)
        assert (ds[data_column_names]).equals(ds_ff[data_column_names])


if __name__ == "__main__":
    tester = unittest.main()



"""
Benchmarks for 'grouping' functions.
"""
__all__ = ["bench_groupbyhash", "bench_groupbylex", "bench_groupbypack"]

import itertools
import logging
from typing import List

import numpy as np
from numpy.random import default_rng

from ..rt_dataset import Dataset
from ..rt_grouping import Grouping
from ..rt_numpy import empty, groupbyhash, groupbylex, groupbypack
from .benchmark import _timestamp_funcs
from .rand_data import rand_array, rand_fancyindex, rand_keyarray
from .runner import create_comparison_dataset, create_trial_dataset

logger = logging.getLogger(__name__)
"""The logger for this module."""

timestamper = _timestamp_funcs["get_nano_time"]
"""The timestamping function to use in benchmarks."""


def bench_groupbyhash(**kwargs) -> Dataset:
    # TODO: Add additional dimensions:
    #  * key dtype(s) -- use this to also bench with multiple keys
    #  * provide pcutoffs array?
    #  * number of threads
    #  * recycler on/off

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 1
    iters = 5

    # Parameters we'll sweep over for the benchmark.
    rng_seeds = [12345]
    dtypes = [np.int16, np.int32, np.dtype("S4"), np.dtype("S10"), np.dtype("<U8")]
    key_unique_counts = [
        100,
        10_000,
    ]  # TODO: Maybe also sweep over constant (across rowcounts) key frequency, e.g. 0.1%, 1%, 10%, 20% unique keys -- this would work better for larger rowcounts
    rowcounts = [
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        100_000_000,
    ]  # TODO: Add 1G and 2G -- these need to be optional since smaller machines will run out of memory
    use_hints = [False, True]

    setup_params = itertools.product(rng_seeds, dtypes, key_unique_counts, rowcounts)

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for rng_seed, dtype, key_unique_count, rowcount in setup_params:
        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        key_data = rand_keyarray(rng, rowcount, dtype=np.dtype(dtype), unique_count=key_unique_count)

        # Sweep over other parameters that aren't required by the setup phase.
        other_params = use_hints  # itertools.product(use_hints, ...)
        for use_hint in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    hint_size = key_unique_count if use_hint else 0
                    groupbyhash(key_data, hint_size=hint_size)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "dtype": np.dtype(dtype).name,
                    "key_unique_count": key_unique_count,
                    "rowcount": rowcount,
                    # Other parameters
                    "use_hint": use_hint,
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_groupbylex(**kwargs) -> Dataset:
    # TODO: Add additional dimensions:
    #  * key dtype(s) -- use this to also bench with multiple keys
    #  * provide pcutoffs array?
    #  * number of threads
    #  * recycler on/off

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 1
    iters = 5

    # Parameters we'll sweep over for the benchmark and are required for the setup phase.
    rng_seeds = [12345]
    dtypes = [np.int16, np.int32, np.dtype("S4"), np.dtype("S10"), np.dtype("<U8")]
    key_unique_counts = [
        100,
        10_000,
    ]  # TODO: Maybe also sweep over constant (across rowcounts) key frequency, e.g. 0.1%, 1%, 10%, 20% unique keys -- this would work better for larger rowcounts
    rowcounts = [
        10_000,
        100_000,
        1_000_000,
        10_000_000,
    ]  # , 100_000_000]   # TODO: Add 1G and 2G -- these need to be optional since smaller machines will run out of memory

    setup_params = itertools.product(rng_seeds, dtypes, key_unique_counts, rowcounts)

    # Parameters to sweep over for the benchmark which are *not* required for the setup phase.
    # TODO: Enable this parameter once we support multi-keys in this benchmark.
    as_recarrays = [False]  # , True]     # Use to set the 'rec' parameter for groupbylex; only meaningful for multi-key

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for rng_seed, dtype, key_unique_count, rowcount in setup_params:
        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        key_data = rand_keyarray(rng, rowcount, dtype=np.dtype(dtype), unique_count=key_unique_count)

        # Sweep over other parameters that aren't required by the setup phase.
        for as_recarray in as_recarrays:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    groupbylex(key_data, rec=as_recarray)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "dtype": np.dtype(dtype).name,
                    "key_unique_count": key_unique_count,
                    "rowcount": rowcount,
                    # Other parameters
                    "as_recarray": as_recarray,
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_groupbypack(**kwargs) -> Dataset:
    import riptide_cpp as rc

    # TODO: Add additional dimensions:
    #  * provide pcutoffs array?
    #  * number of threads
    #  * recycler on/off
    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 1
    iters = 5

    # Parameters we'll sweep over for the benchmark.
    rng_seeds = [12345]
    dtypes = [np.int16, np.int32, np.dtype("S4"), np.dtype("S10"), np.dtype("<U8")]
    key_unique_counts = [
        100,
        10_000,
    ]  # TODO: Maybe also sweep over constant (across rowcounts) key frequency, e.g. 0.1%, 1%, 10%, 20% unique keys -- this would work better for larger rowcounts
    rowcounts = [
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        100_000_000,
    ]  # TODO: Add 1G and 2G -- these need to be optional since smaller machines will run out of memory

    setup_params = itertools.product(rng_seeds, dtypes, key_unique_counts, rowcounts)

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for rng_seed, dtype, key_unique_count, rowcount in setup_params:
        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        key_data = rand_keyarray(rng, rowcount, dtype=np.dtype(dtype), unique_count=key_unique_count)
        gb_data = groupbyhash(key_data, hint_size=key_unique_count)

        # Sweep over other parameters that aren't required by the setup phase.
        # (There are no other parameters to sweep over for this benchmark.)
        for _ in [None]:  # Empty dimension for "other" parameters
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    ncountgroup = rc.BinCount(gb_data["iKey"], gb_data["unique_count"] + 1)
                    groupbypack(gb_data["iKey"], ncountgroup)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "dtype": np.dtype(dtype).name,
                    "key_unique_count": key_unique_count,
                    "rowcount": rowcount,
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)



from .groupby_unit_test_parameters import *


def generate(numb_tests):

    PARAMETERS = groupby_parameters()

    def concat_list(lst):
        string = ""
        for s in lst:
            string += "_" + s
        return string

    def str_header():
        return """
        ##                                                             ##
        #                                                               #
        #   THIS TEST WAS AUTOGENERATED BY groupby_test_generator.py    #
        #                                                               #
        ##   
        """

    def str_imports():
        return """\n\nfrom .groupby_unit_test_parameters import *
        \nimport pandas as pd
        \nimport riptable as rt
        \nimport unittest"""

    def str_agg_params():
        return str(PARAMETERS.agg_list)

    def str_params():
        return (
            str(PARAMETERS.val_cols)
            + ", "
            + str(PARAMETERS.key_cols)
            + ", "
            + str(PARAMETERS.symbol_ratio)[0:4]
            + ", "
            + str_agg_params()
        )

    def str_test_class_init():
        return "\ntest_class = groupby_everything(" + str_params() + " )"

    def str_class_header():
        return "\nclass autogenerated_gb_tests(unittest.TestCase):"

    def str_main_tail():
        return '\n\nif __name__ == "__main__": \n\ttester = unittest.main()'

    def str_safe_assert_func():
        return "\n\tdef safe_assert(self, ary1, ary2): \n\t\tfor a, b in zip(ary1, ary2):\n\t\t\tif a == a and b == b: \n\t\t\t\tself.assertAlmostEqual(a,b,places=7)"

    def str_recycle_off():
        return "\n\t\trt.FastArray._ROFF()"

    def str_recycle_on():
        return "\n\t\trt.FastArray._RON()"

    def str_threads_off():
        return "\n\t\trt.FastArray._TOFF()"

    def str_threads_on():
        return "\n\t\trt.FastArray._TON()"

    def str_standard_test():

        s = "\n\tdef test_multikey_" + trait_str() + "(self):"
        s += "\n\t\taggs = " + str_agg_params()
        s += "\n\t\ttest_class = groupby_everything(" + str_params() + " )"
        s += "\n\t\tpd_out = pd.DataFrame(test_class.data).groupby(KEY_COLUMN_NAMES[:test_class.key_columns]).agg(test_class.aggregation_functions)"
        s += "\n\t\tsf_out = rt.Dataset(test_class.data).groupby(KEY_COLUMN_NAMES[:test_class.key_columns]).agg(test_class.aggregation_functions)"

        s += "\n\t\tfor func in aggs:"
        s += "\n\t\t\tfor i in range(0, test_class.val_columns):"
        s += "\n\t\t\t\tcolumn=VAL_COLUMN_NAMES[i]"
        s += "\n\t\t\t\tself.safe_assert(pd_out[column][func], sf_out[func.title()][column])"

        return s

    def str_norecycle_test():
        s = "\n\tdef test_multikey_NORECYCLING" + trait_str() + "(self):"
        s += str_recycle_off()
        s += "\n\t\taggs = " + str_agg_params()
        s += "\n\t\ttest_class = groupby_everything(" + str_params() + " )"
        s += "\n\t\tpd_out = pd.DataFrame(test_class.data).groupby(KEY_COLUMN_NAMES[:test_class.key_columns]).agg(test_class.aggregation_functions)"
        s += "\n\t\tsf_out = rt.Dataset(test_class.data).groupby(KEY_COLUMN_NAMES[:test_class.key_columns]).agg(test_class.aggregation_functions)"

        s += "\n\t\tfor func in aggs:"
        s += "\n\t\t\tfor i in range(0, test_class.val_columns):"
        s += "\n\t\t\t\tcolumn=VAL_COLUMN_NAMES[i]"
        s += "\n\t\t\t\tself.safe_assert(pd_out[column][func], sf_out[func.title()][column])"
        s += str_recycle_on()

        return s

    def str_nothreads_test():
        s = "\n\tdef test_multikey_NOTHREADS" + trait_str() + "(self):"
        s += str_threads_off()
        s += "\n\t\taggs = " + str_agg_params()
        s += "\n\t\ttest_class = groupby_everything(" + str_params() + " )"
        s += "\n\t\tpd_out = pd.DataFrame(test_class.data).groupby(KEY_COLUMN_NAMES[:test_class.key_columns]).agg(test_class.aggregation_functions)"
        s += "\n\t\tsf_out = rt.Dataset(test_class.data).groupby(KEY_COLUMN_NAMES[:test_class.key_columns]).agg(test_class.aggregation_functions)"

        s += "\n\t\tfor func in aggs:"
        s += "\n\t\t\tfor i in range(0, test_class.val_columns):"
        s += "\n\t\t\t\tcolumn=VAL_COLUMN_NAMES[i]"
        s += "\n\t\t\t\tself.safe_assert(pd_out[column][func], sf_out[func.title()][column])"
        s += str_threads_on()

        return s

    def str_norecycle_nothreads_test():
        s = "\n\tdef test_multikey_NORECYCLINGNOTHREADS" + trait_str() + "(self):"
        s += str_threads_off()
        s += str_recycle_off()
        s += "\n\t\taggs = " + str_agg_params()
        s += "\n\t\ttest_class = groupby_everything(" + str_params() + " )"
        s += "\n\t\tpd_out = pd.DataFrame(test_class.data).groupby(KEY_COLUMN_NAMES[:test_class.key_columns]).agg(test_class.aggregation_functions)"
        s += "\n\t\tsf_out = rt.Dataset(test_class.data).groupby(KEY_COLUMN_NAMES[:test_class.key_columns]).agg(test_class.aggregation_functions)"

        s += "\n\t\tfor func in aggs:"
        s += "\n\t\t\tfor i in range(0, test_class.val_columns):"
        s += "\n\t\t\t\tcolumn=VAL_COLUMN_NAMES[i]"
        s += "\n\t\t\t\tself.safe_assert(pd_out[column][func], sf_out[func.title()][column])"
        s += str_threads_on()
        s += str_recycle_on()

        return s

    test_str = str_header()
    test_str += str_imports()
    test_str += str_class_header()
    test_str += str_safe_assert_func()

    for i in range(0, numb_tests):
        aggs = "__aggs" + concat_list(PARAMETERS.agg_list)
        symbs = "__symb_ratio_" + str(PARAMETERS.symbol_ratio).replace(".", "")[0:4]
        val_cols = "__nvalcols_" + str(PARAMETERS.val_cols)
        key_cols = "__nkeycols_" + str(PARAMETERS.key_cols)

        def trait_str():
            return aggs + symbs + val_cols + key_cols

        s = ""
        s += str_standard_test()
        # s += str_norecycle_test()
        # s += str_nothreads_test()
        # s += str_norecycle_nothreads_test()

        PARAMETERS.update()
        test_str += s
        print(s)

    test_str += str_main_tail()

    filename = "test_sfw_groupby_autotest_aggregated_functions.py"
    file = open(filename, "w")
    file.write(test_str)
    file.close()


if __name__ == "__main__":
    generate(200)
